# Code generated. DO NOT EDIT.
# LSP v3.17.0
# NOTE: https://pypi.org/project/ts2python/ 기반 생성 전환은 추후 검토 대상으로 남긴다.

"""
This file provides the python interface corresponding to the requests and notifications defined in Typescript in the language server protocol.
This file is obtained from https://github.com/predragnikolic/OLSP under the MIT License with the following terms:

MIT License

Copyright (c) 2023 Предраг Николић

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections.abc import Awaitable, Callable
from typing import Union

from solidlsp.lsp_protocol_handler import lsp_types


SendRequestCallable = Callable[[str, object], Awaitable[object]]
SendNotificationCallable = Callable[[str, object], object | None]


class LspRequest:
    def __init__(self, send_request: SendRequestCallable) -> None:
        self.send_request = send_request

    async def implementation(
        self, params: lsp_types.ImplementationParams
    ) -> Union["lsp_types.Definition", list["lsp_types.LocationLink"], None]:
        """A request to resolve the implementation locations of a symbol at a given text
        document position. The request's parameter is of type [TextDocumentPositionParams]
        (#TextDocumentPositionParams) the response is of type {@link Definition} or a
        Thenable that resolves to such.
        """
        return await self.send_request("textDocument/implementation", params)

    async def type_definition(
        self, params: lsp_types.TypeDefinitionParams
    ) -> Union["lsp_types.Definition", list["lsp_types.LocationLink"], None]:
        """A request to resolve the type definition locations of a symbol at a given text
        document position. The request's parameter is of type [TextDocumentPositionParams]
        (#TextDocumentPositionParams) the response is of type {@link Definition} or a
        Thenable that resolves to such.
        """
        return await self.send_request("textDocument/typeDefinition", params)

    async def document_color(self, params: lsp_types.DocumentColorParams) -> list["lsp_types.ColorInformation"]:
        """A request to list all color symbols found in a given text document. The request's
        parameter is of type {@link DocumentColorParams} the
        response is of type {@link ColorInformation ColorInformation[]} or a Thenable
        that resolves to such.
        """
        return await self.send_request("textDocument/documentColor", params)

    async def color_presentation(self, params: lsp_types.ColorPresentationParams) -> list["lsp_types.ColorPresentation"]:
        """A request to list all presentation for a color. The request's
        parameter is of type {@link ColorPresentationParams} the
        response is of type {@link ColorInformation ColorInformation[]} or a Thenable
        that resolves to such.
        """
        return await self.send_request("textDocument/colorPresentation", params)

    async def folding_range(self, params: lsp_types.FoldingRangeParams) -> list["lsp_types.FoldingRange"] | None:
        """A request to provide folding ranges in a document. The request's
        parameter is of type {@link FoldingRangeParams}, the
        response is of type {@link FoldingRangeList} or a Thenable
        that resolves to such.
        """
        return await self.send_request("textDocument/foldingRange", params)

    async def declaration(
        self, params: lsp_types.DeclarationParams
    ) -> Union["lsp_types.Declaration", list["lsp_types.LocationLink"], None]:
        """A request to resolve the type definition locations of a symbol at a given text
        document position. The request's parameter is of type [TextDocumentPositionParams]
        (#TextDocumentPositionParams) the response is of type {@link Declaration}
        or a typed array of {@link DeclarationLink} or a Thenable that resolves
        to such.
        """
        return await self.send_request("textDocument/declaration", params)

    async def selection_range(self, params: lsp_types.SelectionRangeParams) -> list["lsp_types.SelectionRange"] | None:
        """A request to provide selection ranges in a document. The request's
        parameter is of type {@link SelectionRangeParams}, the
        response is of type {@link SelectionRange SelectionRange[]} or a Thenable
        that resolves to such.
        """
        return await self.send_request("textDocument/selectionRange", params)

    async def prepare_call_hierarchy(self, params: lsp_types.CallHierarchyPrepareParams) -> list["lsp_types.CallHierarchyItem"] | None:
        """A request to result a `CallHierarchyItem` in a document at a given position.
        Can be used as an input to an incoming or outgoing call hierarchy.

        @since 3.16.0
        """
        return await self.send_request("textDocument/prepareCallHierarchy", params)

    async def incoming_calls(
        self, params: lsp_types.CallHierarchyIncomingCallsParams
    ) -> list["lsp_types.CallHierarchyIncomingCall"] | None:
        """A request to resolve the incoming calls for a given `CallHierarchyItem`.

        @since 3.16.0
        """
        return await self.send_request("callHierarchy/incomingCalls", params)

    async def outgoing_calls(
        self, params: lsp_types.CallHierarchyOutgoingCallsParams
    ) -> list["lsp_types.CallHierarchyOutgoingCall"] | None:
        """A request to resolve the outgoing calls for a given `CallHierarchyItem`.

        @since 3.16.0
        """
        return await self.send_request("callHierarchy/outgoingCalls", params)

    async def semantic_tokens_full(self, params: lsp_types.SemanticTokensParams) -> Union["lsp_types.SemanticTokens", None]:
        """@since 3.16.0"""
        return await self.send_request("textDocument/semanticTokens/full", params)

    async def semantic_tokens_delta(
        self, params: lsp_types.SemanticTokensDeltaParams
    ) -> Union["lsp_types.SemanticTokens", "lsp_types.SemanticTokensDelta", None]:
        """@since 3.16.0"""
        return await self.send_request("textDocument/semanticTokens/full/delta", params)

    async def semantic_tokens_range(self, params: lsp_types.SemanticTokensRangeParams) -> Union["lsp_types.SemanticTokens", None]:
        """@since 3.16.0"""
        return await self.send_request("textDocument/semanticTokens/range", params)

    async def linked_editing_range(self, params: lsp_types.LinkedEditingRangeParams) -> Union["lsp_types.LinkedEditingRanges", None]:
        """A request to provide ranges that can be edited together.

        @since 3.16.0
        """
        return await self.send_request("textDocument/linkedEditingRange", params)

    async def will_create_files(self, params: lsp_types.CreateFilesParams) -> Union["lsp_types.WorkspaceEdit", None]:
        """The will create files request is sent from the client to the server before files are actually
        created as long as the creation is triggered from within the client.

        @since 3.16.0
        """
        return await self.send_request("workspace/willCreateFiles", params)

    async def will_rename_files(self, params: lsp_types.RenameFilesParams) -> Union["lsp_types.WorkspaceEdit", None]:
        """The will rename files request is sent from the client to the server before files are actually
        renamed as long as the rename is triggered from within the client.

        @since 3.16.0
        """
        return await self.send_request("workspace/willRenameFiles", params)

    async def will_delete_files(self, params: lsp_types.DeleteFilesParams) -> Union["lsp_types.WorkspaceEdit", None]:
        """The did delete files notification is sent from the client to the server when
        files were deleted from within the client.

        @since 3.16.0
        """
        return await self.send_request("workspace/willDeleteFiles", params)

    async def moniker(self, params: lsp_types.MonikerParams) -> list["lsp_types.Moniker"] | None:
        """A request to get the moniker of a symbol at a given text document position.
        The request parameter is of type {@link TextDocumentPositionParams}.
        The response is of type {@link Moniker Moniker[]} or `null`.
        """
        return await self.send_request("textDocument/moniker", params)

    async def prepare_type_hierarchy(self, params: lsp_types.TypeHierarchyPrepareParams) -> list["lsp_types.TypeHierarchyItem"] | None:
        """A request to result a `TypeHierarchyItem` in a document at a given position.
        Can be used as an input to a subtypes or supertypes type hierarchy.

        @since 3.17.0
        """
        return await self.send_request("textDocument/prepareTypeHierarchy", params)

    async def type_hierarchy_supertypes(
        self, params: lsp_types.TypeHierarchySupertypesParams
    ) -> list["lsp_types.TypeHierarchyItem"] | None:
        """A request to resolve the supertypes for a given `TypeHierarchyItem`.

        @since 3.17.0
        """
        return await self.send_request("typeHierarchy/supertypes", params)

    async def type_hierarchy_subtypes(self, params: lsp_types.TypeHierarchySubtypesParams) -> list["lsp_types.TypeHierarchyItem"] | None:
        """A request to resolve the subtypes for a given `TypeHierarchyItem`.

        @since 3.17.0
        """
        return await self.send_request("typeHierarchy/subtypes", params)

    async def inline_value(self, params: lsp_types.InlineValueParams) -> list["lsp_types.InlineValue"] | None:
        """A request to provide inline values in a document. The request's parameter is of
        type {@link InlineValueParams}, the response is of type
        {@link InlineValue InlineValue[]} or a Thenable that resolves to such.

        @since 3.17.0
        """
        return await self.send_request("textDocument/inlineValue", params)

    async def inlay_hint(self, params: lsp_types.InlayHintParams) -> list["lsp_types.InlayHint"] | None:
        """A request to provide inlay hints in a document. The request's parameter is of
        type {@link InlayHintsParams}, the response is of type
        {@link InlayHint InlayHint[]} or a Thenable that resolves to such.

        @since 3.17.0
        """
        return await self.send_request("textDocument/inlayHint", params)

    async def resolve_inlay_hint(self, params: lsp_types.InlayHint) -> "lsp_types.InlayHint":
        """A request to resolve additional properties for an inlay hint.
        The request's parameter is of type {@link InlayHint}, the response is
        of type {@link InlayHint} or a Thenable that resolves to such.

        @since 3.17.0
        """
        return await self.send_request("inlayHint/resolve", params)

    async def text_document_diagnostic(self, params: lsp_types.DocumentDiagnosticParams) -> "lsp_types.DocumentDiagnosticReport":
        """The document diagnostic request definition.

        @since 3.17.0
        """
        return await self.send_request("textDocument/diagnostic", params)

    async def workspace_diagnostic(self, params: lsp_types.WorkspaceDiagnosticParams) -> "lsp_types.WorkspaceDiagnosticReport":
        """The workspace diagnostic request definition.

        @since 3.17.0
        """
        return await self.send_request("workspace/diagnostic", params)

    async def initialize(self, params: lsp_types.InitializeParams) -> "lsp_types.InitializeResult":
        """The initialize request is sent from the client to the server.
        It is sent once as the request after starting up the server.
        The requests parameter is of type {@link InitializeParams}
        the response if of type {@link InitializeResult} of a Thenable that
        resolves to such.
        """
        return await self.send_request("initialize", params)

    async def shutdown(self) -> None:
        """A shutdown request is sent from the client to the server.
        It is sent once when the client decides to shutdown the
        server. The only notification that is sent after a shutdown request
        is the exit event.
        """
        return await self.send_request("shutdown")

    async def will_save_wait_until(self, params: lsp_types.WillSaveTextDocumentParams) -> list["lsp_types.TextEdit"] | None:
        """A document will save request is sent from the client to the server before
        the document is actually saved. The request can return an array of TextEdits
        which will be applied to the text document before it is saved. Please note that
        clients might drop results if computing the text edits took too long or if a
        server constantly fails on this request. This is done to keep the save fast and
        reliable.
        """
        return await self.send_request("textDocument/willSaveWaitUntil", params)

    async def completion(
        self, params: lsp_types.CompletionParams
    ) -> Union[list["lsp_types.CompletionItem"], "lsp_types.CompletionList", None]:
        """Request to request completion at a given text document position. The request's
        parameter is of type {@link TextDocumentPosition} the response
        is of type {@link CompletionItem CompletionItem[]} or {@link CompletionList}
        or a Thenable that resolves to such.

        The request can delay the computation of the {@link CompletionItem.detail `detail`}
        and {@link CompletionItem.documentation `documentation`} properties to the `completionItem/resolve`
        request. However, properties that are needed for the initial sorting and filtering, like `sortText`,
        `filterText`, `insertText`, and `textEdit`, must not be changed during resolve.
        """
        return await self.send_request("textDocument/completion", params)

    async def resolve_completion_item(self, params: lsp_types.CompletionItem) -> "lsp_types.CompletionItem":
        """Request to resolve additional information for a given completion item.The request's
        parameter is of type {@link CompletionItem} the response
        is of type {@link CompletionItem} or a Thenable that resolves to such.
        """
        return await self.send_request("completionItem/resolve", params)

    async def hover(self, params: lsp_types.HoverParams) -> Union["lsp_types.Hover", None]:
        """Request to request hover information at a given text document position. The request's
        parameter is of type {@link TextDocumentPosition} the response is of
        type {@link Hover} or a Thenable that resolves to such.
        """
        return await self.send_request("textDocument/hover", params)

    async def signature_help(self, params: lsp_types.SignatureHelpParams) -> Union["lsp_types.SignatureHelp", None]:
        return await self.send_request("textDocument/signatureHelp", params)

    async def definition(self, params: lsp_types.DefinitionParams) -> Union["lsp_types.Definition", list["lsp_types.LocationLink"], None]:
        """A request to resolve the definition location of a symbol at a given text
        document position. The request's parameter is of type [TextDocumentPosition]
        (#TextDocumentPosition) the response is of either type {@link Definition}
        or a typed array of {@link DefinitionLink} or a Thenable that resolves
        to such.
        """
        return await self.send_request("textDocument/definition", params)

    async def references(self, params: lsp_types.ReferenceParams) -> list["lsp_types.Location"] | None:
        """A request to resolve project-wide references for the symbol denoted
        by the given text document position. The request's parameter is of
        type {@link ReferenceParams} the response is of type
        {@link Location Location[]} or a Thenable that resolves to such.
        """
        return await self.send_request("textDocument/references", params)

    async def document_highlight(self, params: lsp_types.DocumentHighlightParams) -> list["lsp_types.DocumentHighlight"] | None:
        """Request to resolve a {@link DocumentHighlight} for a given
        text document position. The request's parameter is of type [TextDocumentPosition]
        (#TextDocumentPosition) the request response is of type [DocumentHighlight[]]
        (#DocumentHighlight) or a Thenable that resolves to such.
        """
        return await self.send_request("textDocument/documentHighlight", params)

    async def document_symbol(
        self, params: lsp_types.DocumentSymbolParams
    ) -> list["lsp_types.SymbolInformation"] | list["lsp_types.DocumentSymbol"] | None:
        """A request to list all symbols found in a given text document. The request's
        parameter is of type {@link TextDocumentIdentifier} the
        response is of type {@link SymbolInformation SymbolInformation[]} or a Thenable
        that resolves to such.
        """
        return await self.send_request("textDocument/documentSymbol", params)

    async def code_action(self, params: lsp_types.CodeActionParams) -> list[Union["lsp_types.Command", "lsp_types.CodeAction"]] | None:
        """A request to provide commands for the given text document and range."""
        return await self.send_request("textDocument/codeAction", params)

    async def resolve_code_action(self, params: lsp_types.CodeAction) -> "lsp_types.CodeAction":
        """Request to resolve additional information for a given code action.The request's
        parameter is of type {@link CodeAction} the response
        is of type {@link CodeAction} or a Thenable that resolves to such.
        """
        return await self.send_request("codeAction/resolve", params)

    async def workspace_symbol(
        self, params: lsp_types.WorkspaceSymbolParams
    ) -> list["lsp_types.SymbolInformation"] | list["lsp_types.WorkspaceSymbol"] | None:
        """A request to list project-wide symbols matching the query string given
        by the {@link WorkspaceSymbolParams}. The response is
        of type {@link SymbolInformation SymbolInformation[]} or a Thenable that
        resolves to such.

        @since 3.17.0 - support for WorkspaceSymbol in the returned data. Clients
         need to advertise support for WorkspaceSymbols via the client capability
         `workspace.symbol.resolveSupport`.
        """
        return await self.send_request("workspace/symbol", params)

    async def resolve_workspace_symbol(self, params: lsp_types.WorkspaceSymbol) -> "lsp_types.WorkspaceSymbol":
        """A request to resolve the range inside the workspace
        symbol's location.

        @since 3.17.0
        """
        return await self.send_request("workspaceSymbol/resolve", params)

    async def code_lens(self, params: lsp_types.CodeLensParams) -> list["lsp_types.CodeLens"] | None:
        """A request to provide code lens for the given text document."""
        return await self.send_request("textDocument/codeLens", params)

    async def resolve_code_lens(self, params: lsp_types.CodeLens) -> "lsp_types.CodeLens":
        """A request to resolve a command for a given code lens."""
        return await self.send_request("codeLens/resolve", params)

    async def document_link(self, params: lsp_types.DocumentLinkParams) -> list["lsp_types.DocumentLink"] | None:
        """A request to provide document links"""
        return await self.send_request("textDocument/documentLink", params)

    async def resolve_document_link(self, params: lsp_types.DocumentLink) -> "lsp_types.DocumentLink":
        """Request to resolve additional information for a given document link. The request's
        parameter is of type {@link DocumentLink} the response
        is of type {@link DocumentLink} or a Thenable that resolves to such.
        """
        return await self.send_request("documentLink/resolve", params)

    async def formatting(self, params: lsp_types.DocumentFormattingParams) -> list["lsp_types.TextEdit"] | None:
        """A request to to format a whole document."""
        return await self.send_request("textDocument/formatting", params)

    async def range_formatting(self, params: lsp_types.DocumentRangeFormattingParams) -> list["lsp_types.TextEdit"] | None:
        """A request to to format a range in a document."""
        return await self.send_request("textDocument/rangeFormatting", params)

    async def on_type_formatting(self, params: lsp_types.DocumentOnTypeFormattingParams) -> list["lsp_types.TextEdit"] | None:
        """A request to format a document on type."""
        return await self.send_request("textDocument/onTypeFormatting", params)

    async def rename(self, params: lsp_types.RenameParams) -> Union["lsp_types.WorkspaceEdit", None]:
        """A request to rename a symbol."""
        return await self.send_request("textDocument/rename", params)

    async def prepare_rename(self, params: lsp_types.PrepareRenameParams) -> Union["lsp_types.PrepareRenameResult", None]:
        """A request to test and perform the setup necessary for a rename.

        @since 3.16 - support for default behavior
        """
        return await self.send_request("textDocument/prepareRename", params)

    async def execute_command(self, params: lsp_types.ExecuteCommandParams) -> Union["lsp_types.LSPAny", None]:
        """A request send from the client to the server to execute a command. The request might return
        a workspace edit which the client will apply to the workspace.
        """
        return await self.send_request("workspace/executeCommand", params)


class LspNotification:
    def __init__(self, send_notification: SendNotificationCallable) -> None:
        self.send_notification = send_notification

    def did_change_workspace_folders(self, params: lsp_types.DidChangeWorkspaceFoldersParams) -> None:
        """The `workspace/didChangeWorkspaceFolders` notification is sent from the client to the server when the workspace
        folder configuration changes.
        """
        return self.send_notification("workspace/didChangeWorkspaceFolders", params)

    def cancel_work_done_progress(self, params: lsp_types.WorkDoneProgressCancelParams) -> None:
        """The `window/workDoneProgress/cancel` notification is sent from  the client to the server to cancel a progress
        initiated on the server side.
        """
        return self.send_notification("window/workDoneProgress/cancel", params)

    def did_create_files(self, params: lsp_types.CreateFilesParams) -> None:
        """The did create files notification is sent from the client to the server when
        files were created from within the client.

        @since 3.16.0
        """
        return self.send_notification("workspace/didCreateFiles", params)

    def did_rename_files(self, params: lsp_types.RenameFilesParams) -> None:
        """The did rename files notification is sent from the client to the server when
        files were renamed from within the client.

        @since 3.16.0
        """
        return self.send_notification("workspace/didRenameFiles", params)

    def did_delete_files(self, params: lsp_types.DeleteFilesParams) -> None:
        """The will delete files request is sent from the client to the server before files are actually
        deleted as long as the deletion is triggered from within the client.

        @since 3.16.0
        """
        return self.send_notification("workspace/didDeleteFiles", params)

    def did_open_notebook_document(self, params: lsp_types.DidOpenNotebookDocumentParams) -> None:
        """A notification sent when a notebook opens.

        @since 3.17.0
        """
        return self.send_notification("notebookDocument/didOpen", params)

    def did_change_notebook_document(self, params: lsp_types.DidChangeNotebookDocumentParams) -> None:
        return self.send_notification("notebookDocument/didChange", params)

    def did_save_notebook_document(self, params: lsp_types.DidSaveNotebookDocumentParams) -> None:
        """A notification sent when a notebook document is saved.

        @since 3.17.0
        """
        return self.send_notification("notebookDocument/didSave", params)

    def did_close_notebook_document(self, params: lsp_types.DidCloseNotebookDocumentParams) -> None:
        """A notification sent when a notebook closes.

        @since 3.17.0
        """
        return self.send_notification("notebookDocument/didClose", params)

    def initialized(self, params: lsp_types.InitializedParams) -> None:
        """The initialized notification is sent from the client to the
        server after the client is fully initialized and the server
        is allowed to send requests from the server to the client.
        """
        return self.send_notification("initialized", params)

    def exit(self) -> None:
        """The exit event is sent from the client to the server to
        ask the server to exit its process.
        """
        return self.send_notification("exit")

    def workspace_did_change_configuration(self, params: lsp_types.DidChangeConfigurationParams) -> None:
        """The configuration change notification is sent from the client to the server
        when the client's configuration has changed. The notification contains
        the changed configuration as defined by the language client.
        """
        return self.send_notification("workspace/didChangeConfiguration", params)

    def did_open_text_document(self, params: lsp_types.DidOpenTextDocumentParams) -> None:
        """The document open notification is sent from the client to the server to signal
        newly opened text documents. The document's truth is now managed by the client
        and the server must not try to read the document's truth using the document's
        uri. Open in this sense means it is managed by the client. It doesn't necessarily
        mean that its content is presented in an editor. An open notification must not
        be sent more than once without a corresponding close notification send before.
        This means open and close notification must be balanced and the max open count
        is one.
        """
        return self.send_notification("textDocument/didOpen", params)

    def did_change_text_document(self, params: lsp_types.DidChangeTextDocumentParams) -> None:
        """The document change notification is sent from the client to the server to signal
        changes to a text document.
        """
        return self.send_notification("textDocument/didChange", params)

    def did_close_text_document(self, params: lsp_types.DidCloseTextDocumentParams) -> None:
        """The document close notification is sent from the client to the server when
        the document got closed in the client. The document's truth now exists where
        the document's uri points to (e.g. if the document's uri is a file uri the
        truth now exists on disk). As with the open notification the close notification
        is about managing the document's content. Receiving a close notification
        doesn't mean that the document was open in an editor before. A close
        notification requires a previous open notification to be sent.
        """
        return self.send_notification("textDocument/didClose", params)

    def did_save_text_document(self, params: lsp_types.DidSaveTextDocumentParams) -> None:
        """The document save notification is sent from the client to the server when
        the document got saved in the client.
        """
        return self.send_notification("textDocument/didSave", params)

    def will_save_text_document(self, params: lsp_types.WillSaveTextDocumentParams) -> None:
        """A document will save notification is sent from the client to the server before
        the document is actually saved.
        """
        return self.send_notification("textDocument/willSave", params)

    def did_change_watched_files(self, params: lsp_types.DidChangeWatchedFilesParams) -> None:
        """The watched files notification is sent from the client to the server when
        the client detects changes to file watched by the language client.
        """
        return self.send_notification("workspace/didChangeWatchedFiles", params)

    def set_trace(self, params: lsp_types.SetTraceParams) -> None:
        return self.send_notification("$/setTrace", params)

    def cancel_request(self, params: lsp_types.CancelParams) -> None:
        return self.send_notification("$/cancelRequest", params)

    def progress(self, params: lsp_types.ProgressParams) -> None:
        return self.send_notification("$/progress", params)



"""MCP 요청/응답 계약 타입을 정의한다."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class McpError:
    """MCP 오류 객체를 표현한다."""

    code: int
    message: str

    def to_dict(self) -> dict[str, object]:
        """오류 객체를 직렬화 가능한 딕셔너리로 변환한다."""
        return {"code": self.code, "message": self.message}


@dataclass(frozen=True)
class McpResponse:
    """MCP JSON-RPC 응답 객체를 표현한다."""

    request_id: object | None
    result: dict[str, object] | None
    error: McpError | None

    def to_dict(self) -> dict[str, object]:
        """응답을 JSON 직렬화 가능한 딕셔너리로 변환한다."""
        base: dict[str, object] = {"jsonrpc": "2.0", "id": self.request_id}
        if self.error is not None:
            base["error"] = self.error.to_dict()
            return base
        base["result"] = self.result if self.result is not None else {}
        return base



import platform
import subprocess


def subprocess_kwargs() -> dict:
    """
    Returns a dictionary of keyword arguments for subprocess calls, adding platform-specific
    flags that we want to use consistently.
    """
    kwargs = {}
    if platform.system() == "Windows":
        kwargs["creationflags"] = subprocess.CREATE_NO_WINDOW  # type: ignore
    return kwargs


def quote_arg(arg: str) -> str:
    """
    Adds quotes around an argument if it contains spaces.
    """
    if " " not in arg:
        return arg
    return f'"{arg}"'



"""MCP 파이프라인 벤치마크 도구를 제공한다."""

from __future__ import annotations

from sari.core.exceptions import BenchmarkError
from sari.core.models import ErrorResponseDTO
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.mcp.tools.arg_parser import parse_boolean, parse_string_list
from sari.mcp.tools.admin_tools import validate_repo_argument
from sari.mcp.tools.pack1 import Pack1MetaDTO, pack1_error, pack1_success
from sari.services.pipeline_benchmark_service import PipelineBenchmarkService


def _parse_language_filter(arguments: dict[str, object]) -> tuple[tuple[str, ...] | None, ErrorResponseDTO | None]:
    """MCP arguments에서 language_filter를 파싱한다."""
    return parse_string_list(arguments=arguments, key="language_filter")


def _parse_per_language_report(arguments: dict[str, object]) -> tuple[bool, ErrorResponseDTO | None]:
    """MCP arguments에서 per_language_report 플래그를 파싱한다."""
    return parse_boolean(arguments=arguments, key="per_language_report", default=False)


def _benchmark_error(exc: BenchmarkError) -> dict[str, object]:
    """벤치마크 예외를 pack1 오류 응답으로 변환한다."""
    return pack1_error(ErrorResponseDTO(code=exc.context.code, message=exc.context.message))


class PipelineBenchmarkRunTool:
    """pipeline_benchmark_run MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, benchmark_service: PipelineBenchmarkService) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._benchmark_service = benchmark_service

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """벤치마크 실행 결과를 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)
        repo = str(arguments["repo"])
        target_raw = arguments.get("target_files", 50_000)
        profile_raw = arguments.get("profile", "default")
        if not isinstance(target_raw, int):
            return pack1_error(ErrorResponseDTO(code="ERR_INVALID_TARGET_FILES", message="target_files must be integer"))
        if not isinstance(profile_raw, str) or profile_raw.strip() == "":
            return pack1_error(ErrorResponseDTO(code="ERR_INVALID_PROFILE", message="profile must be non-empty string"))
        language_filter, language_filter_error = _parse_language_filter(arguments)
        if language_filter_error is not None:
            return pack1_error(language_filter_error)
        per_language_report, per_language_report_error = _parse_per_language_report(arguments)
        if per_language_report_error is not None:
            return pack1_error(per_language_report_error)
        try:
            summary = self._benchmark_service.run(
                repo_root=repo,
                target_files=target_raw,
                profile=profile_raw,
                language_filter=language_filter,
                per_language_report=per_language_report,
            )
        except BenchmarkError as exc:
            return _benchmark_error(exc)
        return pack1_success(
            {
                "items": [summary],
                "meta": Pack1MetaDTO(
                    candidate_count=1,
                    resolved_count=1,
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )


class PipelineBenchmarkReportTool:
    """pipeline_benchmark_report MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, benchmark_service: PipelineBenchmarkService) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._benchmark_service = benchmark_service

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """최신 벤치마크 결과를 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)
        try:
            summary = self._benchmark_service.get_latest_report()
        except BenchmarkError as exc:
            return _benchmark_error(exc)
        return pack1_success(
            {
                "items": [summary],
                "meta": Pack1MetaDTO(
                    candidate_count=1,
                    resolved_count=1,
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )



"""Erlang Language Server implementation using Erlang LS."""

import logging
import os
import shutil
import subprocess
import threading
import time

from overrides import override

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)


class ErlangLanguageServer(SolidLanguageServer):
    """Language server for Erlang using Erlang LS."""

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates an ErlangLanguageServer instance. This class is not meant to be instantiated directly.
        Use LanguageServer.create() instead.
        """
        self.erlang_ls_path = shutil.which("erlang_ls")
        if not self.erlang_ls_path:
            raise RuntimeError("Erlang LS not found. Install from: https://github.com/erlang-ls/erlang_ls")

        if not self._check_erlang_installation():
            raise RuntimeError("Erlang/OTP not found. Install from: https://www.erlang.org/downloads")

        super().__init__(
            config,
            repository_root_path,
            ProcessLaunchInfo(cmd=[self.erlang_ls_path, "--transport", "stdio"], cwd=repository_root_path),
            "erlang",
            solidlsp_settings,
        )

        # Add server readiness tracking like Elixir
        self.server_ready = threading.Event()

        # Set generous timeout for Erlang LS initialization
        self.set_request_timeout(120.0)

    def _check_erlang_installation(self) -> bool:
        """Check if Erlang/OTP is available."""
        try:
            result = subprocess.run(["erl", "-version"], check=False, capture_output=True, text=True, timeout=10)
            return result.returncode == 0
        except (subprocess.SubprocessError, FileNotFoundError):
            return False

    @classmethod
    def _get_erlang_version(cls) -> str | None:
        """Get the installed Erlang/OTP version or None if not found."""
        try:
            result = subprocess.run(["erl", "-version"], check=False, capture_output=True, text=True, timeout=10)
            if result.returncode == 0:
                return result.stderr.strip()  # erl -version outputs to stderr
        except (subprocess.SubprocessError, FileNotFoundError):
            return None
        return None

    @classmethod
    def _check_rebar3_available(cls) -> bool:
        """Check if rebar3 build tool is available."""
        try:
            result = subprocess.run(["rebar3", "version"], check=False, capture_output=True, text=True, timeout=10)
            return result.returncode == 0
        except (subprocess.SubprocessError, FileNotFoundError):
            return False

    def _start_server(self) -> None:
        """Start Erlang LS server process with proper initialization waiting."""

        def register_capability_handler(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            """Handle window/logMessage notifications from Erlang LS"""
            message_text = msg.get("message", "")
            log.info(f"LSP: window/logMessage: {message_text}")

            # Look for Erlang LS readiness signals
            # Common patterns: "Started Erlang LS", "initialized", "ready"
            readiness_signals = [
                "Started Erlang LS",
                "server started",
                "initialized",
                "ready to serve requests",
                "compilation finished",
                "indexing complete",
            ]

            message_lower = message_text.lower()
            for signal in readiness_signals:
                if signal.lower() in message_lower:
                    log.info(f"Erlang LS readiness signal detected: {message_text}")
                    self.server_ready.set()
                    break

        def do_nothing(params: dict) -> None:
            return

        def check_server_ready(params: dict) -> None:
            """Handle $/progress notifications from Erlang LS as fallback."""
            value = params.get("value", {})

            # Check for initialization completion progress
            if value.get("kind") == "end":
                message = value.get("message", "")
                if any(word in message.lower() for word in ["initialized", "ready", "complete"]):
                    log.info("Erlang LS initialization progress completed")
                    # Set as fallback if no window/logMessage was received
                    if not self.server_ready.is_set():
                        self.server_ready.set()

        # Set up notification handlers
        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_notification("$/progress", check_server_ready)
        self.server.on_notification("window/workDoneProgress/create", do_nothing)
        self.server.on_notification("$/workDoneProgress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting Erlang LS server process")
        self.server.start()

        # Send initialize request
        initialize_params = {
            "processId": None,
            "rootPath": self.repository_root_path,
            "rootUri": f"file://{self.repository_root_path}",
            "capabilities": {
                "textDocument": {
                    "synchronization": {"didSave": True},
                    "completion": {"dynamicRegistration": True},
                    "definition": {"dynamicRegistration": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {"dynamicRegistration": True},
                    "hover": {"dynamicRegistration": True},
                }
            },
        }

        log.info("Sending initialize request to Erlang LS")
        init_response = self.server.send.initialize(initialize_params)  # type: ignore[arg-type]

        # Verify server capabilities
        if "capabilities" in init_response:
            log.info(f"Erlang LS capabilities: {list(init_response['capabilities'].keys())}")

        self.server.notify.initialized({})

        # Wait for Erlang LS to be ready - adjust timeout based on environment
        is_ci = os.getenv("CI") == "true" or os.getenv("GITHUB_ACTIONS") == "true"
        is_macos = os.uname().sysname == "Darwin" if hasattr(os, "uname") else False

        # macOS in CI can be particularly slow for language server startup
        if is_ci and is_macos:
            ready_timeout = 240.0  # 4 minutes for macOS CI
            env_desc = "macOS CI"
        elif is_ci:
            ready_timeout = 180.0  # 3 minutes for other CI
            env_desc = "CI"
        else:
            ready_timeout = 60.0  # 1 minute for local
            env_desc = "local"

        log.info(f"Waiting up to {ready_timeout} seconds for Erlang LS readiness ({env_desc} environment)...")

        if self.server_ready.wait(timeout=ready_timeout):
            log.info("Erlang LS is ready and available for requests")

            # Add settling period for indexing - adjust based on environment
            settling_time = 15.0 if is_ci else 5.0
            log.info(f"Allowing {settling_time} seconds for Erlang LS indexing to complete...")
            time.sleep(settling_time)
            log.info("Erlang LS settling period complete")
        else:
            # Set ready anyway and continue - Erlang LS might not send explicit ready messages
            log.warning(f"Erlang LS readiness timeout reached after {ready_timeout}s, proceeding anyway (common in CI)")
            self.server_ready.set()

            # Still give some time for basic initialization even without explicit readiness signal
            basic_settling_time = 20.0 if is_ci else 10.0
            log.info(f"Allowing {basic_settling_time} seconds for basic Erlang LS initialization...")
            time.sleep(basic_settling_time)
            log.info("Basic Erlang LS initialization period complete")

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        # For Erlang projects, we should ignore:
        # - _build: rebar3 build artifacts
        # - deps: dependencies
        # - ebin: compiled beam files
        # - .rebar3: rebar3 cache
        # - logs: log files
        # - node_modules: if the project has JavaScript components
        return super().is_ignored_dirname(dirname) or dirname in [
            "_build",
            "deps",
            "ebin",
            ".rebar3",
            "logs",
            "node_modules",
            "_checkouts",
            "cover",
        ]

    def is_ignored_filename(self, filename: str) -> bool:
        """Check if a filename should be ignored."""
        # Ignore compiled BEAM files
        if filename.endswith(".beam"):
            return True
        # Don't ignore Erlang source files, header files, or configuration files
        return False



"""MCP 도구 인자 공통 파서."""

from __future__ import annotations

from sari.core.models import ErrorResponseDTO


def parse_positive_int(arguments: dict[str, object], key: str, default: int) -> tuple[int, ErrorResponseDTO | None]:
    """양수 정수 인자를 파싱한다."""
    raw = arguments.get(key, default)
    if not isinstance(raw, int) or raw <= 0:
        return 0, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be positive integer")
    return raw, None


def parse_non_negative_int(arguments: dict[str, object], key: str, default: int = 0) -> tuple[int, ErrorResponseDTO | None]:
    """0 이상 정수 인자를 파싱한다."""
    raw = arguments.get(key, default)
    if not isinstance(raw, int) or raw < 0:
        return 0, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be non-negative integer")
    return raw, None


def parse_optional_int(arguments: dict[str, object], key: str, default: int | None) -> tuple[int | None, ErrorResponseDTO | None]:
    """옵셔널 정수 인자를 파싱한다."""
    raw = arguments.get(key, default)
    if raw is None:
        return None, None
    if not isinstance(raw, int):
        return None, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be integer or null")
    return raw, None


def parse_optional_loose_int(
    arguments: dict[str, object],
    key: str,
    default: int | None = None,
) -> tuple[int | None, ErrorResponseDTO | None]:
    """옵셔널 정수 인자를 느슨한 규칙(문자열 숫자 허용)으로 파싱한다."""
    raw = arguments.get(key, default)
    if raw is None:
        return None, None
    if isinstance(raw, int):
        return raw, None
    if isinstance(raw, str):
        text = raw.strip()
        if text == "":
            return None, None
        try:
            return int(text), None
        except ValueError:
            return None, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be integer")
    return None, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be integer")


def parse_non_empty_string(arguments: dict[str, object], key: str) -> tuple[str, ErrorResponseDTO | None]:
    """필수 문자열 인자를 파싱한다."""
    raw = arguments.get(key)
    if not isinstance(raw, str) or raw.strip() == "":
        return "", ErrorResponseDTO(code=f"ERR_{key.upper()}_REQUIRED", message=f"{key} is required")
    return raw.strip(), None


def parse_optional_string(arguments: dict[str, object], key: str) -> str | None:
    """옵셔널 문자열 인자를 파싱한다."""
    raw = arguments.get(key)
    if not isinstance(raw, str):
        return None
    value = raw.strip()
    if value == "":
        return None
    return value


def parse_boolean(arguments: dict[str, object], key: str, default: bool = False) -> tuple[bool, ErrorResponseDTO | None]:
    """불리언/문자열 불리언 인자를 파싱한다."""
    raw = arguments.get(key, default)
    if isinstance(raw, bool):
        return raw, None
    if isinstance(raw, str):
        normalized = raw.strip().lower()
        if normalized in {"1", "true", "yes", "on"}:
            return True, None
        if normalized in {"0", "false", "no", "off"}:
            return False, None
    return False, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be boolean")


def parse_optional_boolean(arguments: dict[str, object], key: str) -> tuple[bool | None, ErrorResponseDTO | None]:
    """옵셔널 불리언 인자를 파싱한다."""
    raw = arguments.get(key)
    if raw is None:
        return None, None
    if isinstance(raw, bool):
        return raw, None
    if isinstance(raw, str):
        normalized = raw.strip().lower()
        if normalized == "":
            return None, None
        if normalized in {"1", "true", "yes", "on"}:
            return True, None
        if normalized in {"0", "false", "no", "off"}:
            return False, None
    return None, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be boolean")


def parse_string_list(
    arguments: dict[str, object],
    key: str,
) -> tuple[tuple[str, ...] | None, ErrorResponseDTO | None]:
    """문자열 또는 문자열 배열을 문자열 튜플로 변환한다."""
    raw = arguments.get(key)
    if raw is None:
        return None, None
    if isinstance(raw, str):
        tokens = [token.strip() for token in raw.split(",") if token.strip() != ""]
        return (None if len(tokens) == 0 else tuple(tokens)), None
    if isinstance(raw, list):
        parsed: list[str] = []
        for item in raw:
            if not isinstance(item, str):
                return None, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} items must be string")
            token = item.strip()
            if token != "":
                parsed.append(token)
        return (None if len(parsed) == 0 else tuple(parsed)), None
    return None, ErrorResponseDTO(code=f"ERR_INVALID_{key.upper()}", message=f"{key} must be string or string[]")



"""데몬 수명주기 서비스를 구현한다."""

from __future__ import annotations

import os
import signal
import socket
import subprocess
import sys
import time
import logging
from typing import TextIO
from datetime import datetime, timedelta, timezone
from pathlib import Path

from sari.core.config import AppConfig
from sari.core.exceptions import DaemonError, ErrorContext
from sari.core.models import DaemonRegistryEntryDTO, DaemonRuntimeDTO, now_iso8601_utc
from sari.db.repositories.daemon_registry_repository import DaemonRegistryRepository
from sari.db.repositories.runtime_repository import RuntimeRepository
from sari.db.repositories.workspace_repository import WorkspaceRepository

log = logging.getLogger(__name__)


class DaemonService:
    """데몬 시작/상태/종료 규칙을 담당한다."""

    def __init__(
        self,
        config: AppConfig,
        runtime_repo: RuntimeRepository,
        workspace_repo: WorkspaceRepository | None = None,
        registry_repo: DaemonRegistryRepository | None = None,
    ) -> None:
        """서비스 생성 시 설정과 저장소를 주입한다."""
        self._config = config
        self._runtime_repo = runtime_repo
        self._workspace_repo = workspace_repo
        self._registry_repo = registry_repo

    def start(self, run_mode: str | None = None) -> DaemonRuntimeDTO:
        """데몬 프로세스를 백그라운드로 시작하고 런타임 상태를 저장한다."""
        self._clear_stale_runtime_if_needed()
        existing = self._runtime_repo.get_runtime()
        if existing is not None and self._is_pid_alive(existing.pid):
            raise DaemonError(ErrorContext(code="ERR_DAEMON_ALREADY_RUNNING", message="이미 데몬이 실행 중입니다"))

        port = self._allocate_port(self._config.preferred_port, self._config.max_port_scan)
        command = [
            sys.executable,
            "-m",
            "sari.daemon_process",
            "--db-path",
            str(self._config.db_path),
            "--host",
            self._config.host,
            "--port",
            str(port),
        ]
        selected_run_mode = self._config.run_mode if run_mode is None else run_mode
        if selected_run_mode not in {"dev", "prod"}:
            raise DaemonError(ErrorContext(code="ERR_INVALID_RUN_MODE", message="run_mode는 dev 또는 prod여야 합니다"))
        command.extend(["--run-mode", selected_run_mode])
        src_root = str(Path(__file__).resolve().parents[2])
        env = os.environ.copy()
        existing_pythonpath = env.get("PYTHONPATH", "")
        env["PYTHONPATH"] = src_root if existing_pythonpath == "" else f"{src_root}:{existing_pythonpath}"
        stdout_stream, stderr_stream = self._open_daemon_log_streams()
        try:
            process = subprocess.Popen(
                command,
                stdout=stdout_stream,
                stderr=stderr_stream,
                start_new_session=True,
                env=env,
            )
        finally:
            # 자식 프로세스가 fd를 복제하므로 부모에서 즉시 닫아도 안전하다.
            stdout_stream.close()
            stderr_stream.close()
        runtime = DaemonRuntimeDTO(
            pid=process.pid,
            host=self._config.host,
            port=port,
            state="running",
            started_at=now_iso8601_utc(),
            session_count=0,
            last_heartbeat_at=now_iso8601_utc(),
            last_exit_reason=None,
        )
        self._runtime_repo.upsert_runtime(runtime)
        self._runtime_repo.reset_session_count()
        self._register_registry_entry(runtime)
        return runtime

    def ensure_running(self, run_mode: str | None = None) -> DaemonRuntimeDTO:
        """데몬이 이미 실행 중이면 attach하고, 아니면 시작한다."""
        self._clear_stale_runtime_if_needed()
        existing = self._runtime_repo.get_runtime()
        if existing is not None and self._is_pid_alive(existing.pid):
            if self._is_attachable_registry_state(existing.pid):
                self._touch_registry(existing.pid)
                return existing
            self.stop()
        return self.start(run_mode=run_mode)

    def status(self) -> DaemonRuntimeDTO | None:
        """현재 데몬 상태를 조회한다."""
        runtime = self._runtime_repo.get_runtime()
        if runtime is None:
            return None
        if self._is_runtime_stale(runtime.last_heartbeat_at):
            self._record_registry_health(pid=runtime.pid, ok=False, error_message="heartbeat stale")
            self._remove_registry_by_pid(runtime.pid)
            self._runtime_repo.clear_runtime()
            return None
        if not self._is_pid_alive(runtime.pid):
            self._record_registry_health(pid=runtime.pid, ok=False, error_message="process dead")
            self._remove_registry_by_pid(runtime.pid)
            self._runtime_repo.clear_runtime()
            return None
        self._touch_registry(runtime.pid)
        return runtime

    def stop(self) -> None:
        """실행 중인 데몬을 종료한다."""
        runtime = self._runtime_repo.get_runtime()
        if runtime is None:
            raise DaemonError(ErrorContext(code="ERR_DAEMON_NOT_RUNNING", message="실행 중인 데몬이 없습니다"))

        try:
            self._signal_process_tree(runtime.pid, signal.SIGTERM)
        except ProcessLookupError as exc:
            self._remove_registry_by_pid(runtime.pid)
            self._runtime_repo.clear_runtime()
            raise DaemonError(ErrorContext(code="ERR_DAEMON_NOT_FOUND", message="데몬 프로세스를 찾을 수 없습니다")) from exc

        deadline = time.time() + self._config.stop_grace_sec
        while time.time() < deadline:
            if not self._is_pid_alive(runtime.pid):
                self._runtime_repo.mark_exit_reason(runtime.pid, "NORMAL_SHUTDOWN", now_iso8601_utc())
                self._remove_registry_by_pid(runtime.pid)
                self._runtime_repo.clear_runtime()
                return
            time.sleep(0.1)

        try:
            self._signal_process_tree(runtime.pid, signal.SIGKILL)
        except ProcessLookupError:
            # 강제 종료 시점에 이미 프로세스가 종료된 경우를 기록한다.
            log.debug("강제 종료 시점에 데몬 프로세스가 이미 종료됨(pid=%s)", runtime.pid)
        self._runtime_repo.mark_exit_reason(runtime.pid, "FORCE_KILLED", now_iso8601_utc())
        self._remove_registry_by_pid(runtime.pid)
        self._runtime_repo.clear_runtime()

    def _allocate_port(self, preferred_port: int, max_scan: int) -> int:
        """사용 가능한 포트를 탐색해 반환한다."""
        for offset in range(max_scan + 1):
            port = preferred_port + offset
            if self._is_port_free(self._config.host, port):
                return port
        raise DaemonError(ErrorContext(code="ERR_PORT_EXHAUSTED", message="사용 가능한 포트를 찾지 못했습니다"))

    def _is_port_free(self, host: str, port: int) -> bool:
        """포트 사용 가능 여부를 검사한다."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                sock.bind((host, port))
            except OSError:
                return False
        return True

    def _is_pid_alive(self, pid: int) -> bool:
        """PID가 살아있는지 확인한다."""
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            return False
        except PermissionError:
            return True
        stat = self._read_process_stat(pid)
        if stat.startswith("Z"):
            return False
        return True

    def _read_process_stat(self, pid: int) -> str:
        """프로세스 상태 문자열(ps stat)을 조회한다."""
        process = subprocess.run(["ps", "-o", "stat=", "-p", str(pid)], capture_output=True, text=True)
        if process.returncode != 0:
            return ""
        return process.stdout.strip()

    def _is_runtime_stale(self, last_heartbeat_at: str) -> bool:
        """heartbeat 기준 stale 상태를 판정한다."""
        try:
            last = datetime.fromisoformat(last_heartbeat_at)
        except ValueError:
            return True
        if last.tzinfo is None:
            last = last.replace(tzinfo=timezone.utc)
        cutoff = datetime.now(timezone.utc) - timedelta(seconds=self._config.daemon_stale_timeout_sec)
        return last < cutoff

    def _clear_stale_runtime_if_needed(self) -> None:
        """stale heartbeat 런타임 레코드를 정리한다."""
        runtime = self._runtime_repo.get_runtime()
        if runtime is None:
            return
        if not self._is_runtime_stale(runtime.last_heartbeat_at):
            return
        if self._is_pid_alive(runtime.pid):
            try:
                self._signal_process_tree(runtime.pid, signal.SIGKILL)
            except ProcessLookupError:
                log.debug("stale runtime 정리 중 이미 프로세스 종료(pid=%s)", runtime.pid)
        self._remove_registry_by_pid(runtime.pid)
        self._runtime_repo.clear_runtime()

    def _signal_process_tree(self, pid: int, sig: signal.Signals) -> None:
        """대상 PID와 같은 프로세스 그룹에 동일 시그널을 전파한다."""
        os.kill(pid, sig)
        try:
            pgid = os.getpgid(pid)
        except ProcessLookupError:
            return
        except OSError:
            return
        try:
            os.killpg(pgid, sig)
        except ProcessLookupError:
            return
        except OSError:
            return

    def _open_daemon_log_streams(self) -> tuple[TextIO, TextIO]:
        """데몬 stdout/stderr 리다이렉트를 위한 로그 파일 스트림을 연다."""
        log_dir = self._config.db_path.parent / "logs"
        try:
            log_dir.mkdir(parents=True, exist_ok=True)
            stdout_stream = (log_dir / "daemon.stdout.log").open(mode="a", encoding="utf-8")
            stderr_stream = (log_dir / "daemon.stderr.log").open(mode="a", encoding="utf-8")
        except OSError as exc:
            raise DaemonError(
                ErrorContext(code="ERR_DAEMON_LOG_OPEN_FAILED", message="데몬 로그 파일을 열지 못했습니다")
            ) from exc
        return stdout_stream, stderr_stream

    def _register_registry_entry(self, runtime: DaemonRuntimeDTO) -> None:
        """런타임 상태를 daemon registry에 등록한다."""
        if self._registry_repo is None:
            return
        workspace_root = self._resolve_registry_workspace_root()
        daemon_id = self._build_daemon_id(runtime)
        entry = DaemonRegistryEntryDTO(
            daemon_id=daemon_id,
            host=runtime.host,
            port=runtime.port,
            pid=runtime.pid,
            workspace_root=workspace_root,
            protocol="http",
            started_at=runtime.started_at,
            last_seen_at=runtime.last_heartbeat_at,
            is_draining=False,
        )
        self._registry_repo.upsert(entry)

    def _touch_registry(self, pid: int) -> None:
        """레지스트리 last_seen을 갱신한다."""
        if self._registry_repo is None:
            return
        daemon_id = self._find_registry_daemon_id_by_pid(pid)
        if daemon_id is None:
            return
        now = now_iso8601_utc()
        self._registry_repo.touch(daemon_id=daemon_id, seen_at=now)
        self._registry_repo.record_health_result(daemon_id=daemon_id, ok=True, health_at=now)

    def _remove_registry_by_pid(self, pid: int) -> None:
        """종료된 PID의 레지스트리 엔트리를 제거한다."""
        if self._registry_repo is None:
            return
        self._registry_repo.remove_by_pid(pid=pid)

    def _record_registry_health(self, pid: int, ok: bool, error_message: str | None = None) -> None:
        """PID 기준으로 레지스트리 헬스 상태를 기록한다."""
        if self._registry_repo is None:
            return
        daemon_id = self._find_registry_daemon_id_by_pid(pid)
        if daemon_id is None:
            return
        self._registry_repo.record_health_result(
            daemon_id=daemon_id,
            ok=ok,
            health_at=now_iso8601_utc(),
            error_message=error_message,
        )

    def _resolve_registry_workspace_root(self) -> str:
        """레지스트리 엔트리에 사용할 워크스페이스 루트를 결정한다."""
        if self._workspace_repo is None:
            return "__global__"
        items = self._workspace_repo.list_all()
        if len(items) == 0:
            return "__global__"
        return items[0].path

    def _build_daemon_id(self, runtime: DaemonRuntimeDTO) -> str:
        """레지스트리용 daemon_id를 구성한다."""
        started_key = runtime.started_at.replace(":", "").replace("-", "")
        return f"daemon-{runtime.pid}-{started_key}"

    def _find_registry_daemon_id_by_pid(self, pid: int) -> str | None:
        """PID로 등록된 daemon_id를 찾는다."""
        if self._registry_repo is None:
            return None
        for item in self._registry_repo.list_all():
            if item.pid == pid:
                return item.daemon_id
        return None

    def _is_attachable_registry_state(self, pid: int) -> bool:
        """현재 PID가 attach 가능한 registry 상태인지 판단한다."""
        if self._registry_repo is None:
            return True
        for item in self._registry_repo.list_all():
            if item.pid != pid:
                continue
            if item.is_draining:
                return False
            if item.deployment_state != "ACTIVE":
                return False
            return True
        return True



# tests/test_workspace_service.py
import pytest
import os
from pathlib import Path
from sari.services.workspace_service import WorkspaceService
from sari.core.exceptions import WorkspaceError
from sari.core.models import Workspace

class MockWorkspaceRepository:
    """테스트를 위한 인메모리 저장소"""
    def __init__(self) -> None:
        self.workspaces: dict[str, Workspace] = {}

    def add(self, workspace: Workspace) -> None:
        self.workspaces[workspace.path] = workspace

    def get_by_path(self, path: str) -> Workspace | None:
        return self.workspaces.get(path)

    def list_all(self) -> list[Workspace]:
        return list(self.workspaces.values())

    def remove(self, path: str) -> None:
        if path in self.workspaces:
            del self.workspaces[path]

@pytest.fixture
def workspace_service() -> WorkspaceService:
    repo = MockWorkspaceRepository()
    return WorkspaceService(repository=repo)

def test_add_workspace_success(workspace_service: WorkspaceService, tmp_path: Path) -> None:
    # Given: 유효한 디렉토리 경로
    path = str(tmp_path / "repo-a")
    os.mkdir(path)
    
    # When: 워크스페이스 추가
    workspace_service.add_workspace(path)
    
    # Then: 등록 확인
    workspaces = workspace_service.list_workspaces()
    assert len(workspaces) == 1
    assert workspaces[0].path == path

def test_add_workspace_non_existent_path(workspace_service: WorkspaceService) -> None:
    # Given: 존재하지 않는 경로
    path = "/non/existent/path"
    
    # When & Then: WorkspaceError 발생 확인
    with pytest.raises(WorkspaceError, match="존재하지 않는 경로입니다"):
        workspace_service.add_workspace(path)

def test_add_workspace_duplicate_path(workspace_service: WorkspaceService, tmp_path: Path) -> None:
    # Given: 이미 등록된 경로
    path = str(tmp_path / "repo-b")
    os.mkdir(path)
    workspace_service.add_workspace(path)
    
    # When & Then: 중복 등록 시 WorkspaceError 발생 확인
    with pytest.raises(WorkspaceError, match="이미 등록된 워크스페이스입니다"):
        workspace_service.add_workspace(path)



"""
This module contains the exceptions raised by the framework.
"""

from solidlsp.ls_config import Language


class SolidLSPException(Exception):
    def __init__(self, message: str, cause: Exception | None = None) -> None:
        """
        Initializes the exception with the given message.

        :param message: the message describing the exception
        :param cause: the original exception that caused this exception, if any.
            For exceptions raised during request handling, this is typically
                * an LSPError for errors returned by the LSP server
                * LanguageServerTerminatedException for errors due to the language server having terminated.
        """
        self.cause = cause
        super().__init__(message)

    def is_language_server_terminated(self) -> bool:
        """
        :return: True if the exception is caused by the language server having terminated as indicated
            by the causing exception being an instance of LanguageServerTerminatedException.
        """
        from .ls_handler import LanguageServerTerminatedException

        return isinstance(self.cause, LanguageServerTerminatedException)

    def get_affected_language(self) -> Language | None:
        """
        :return: the affected language for the case where the exception is caused by the language server having terminated
        """
        from .ls_handler import LanguageServerTerminatedException

        if isinstance(self.cause, LanguageServerTerminatedException):
            return self.cause.language
        return None

    def __str__(self) -> str:
        """
        Returns a string representation of the exception.
        """
        s = super().__str__()
        if self.cause:
            if "\n" in s:
                s += "\n"
            else:
                s += " "
            s += f"(caused by {self.cause})"
        return s



"""관리용 HTTP 엔드포인트를 제공한다."""

from __future__ import annotations

from starlette.responses import JSONResponse

from sari.http.context import HttpContext


async def errors_endpoint(request) -> JSONResponse:
    """데몬 점검 결과와 오류 정보를 반환한다."""
    context: HttpContext = request.app.state.context
    return JSONResponse({"errors": [], "doctor": context.admin_service.doctor()})


async def rescan_endpoint(request) -> JSONResponse:
    """강제 인덱싱 재실행을 요청한다."""
    context: HttpContext = request.app.state.context
    return JSONResponse(context.admin_service.index())


async def repo_candidates_endpoint(request) -> JSONResponse:
    """레포 후보 목록을 반환한다."""
    context: HttpContext = request.app.state.context
    return JSONResponse({"items": context.admin_service.repo_candidates()})


async def doctor_endpoint(request) -> JSONResponse:
    """헬스체크 결과를 반환한다."""
    context: HttpContext = request.app.state.context
    checks = context.admin_service.doctor()
    return JSONResponse({"checks": [{"name": c.name, "passed": c.passed, "detail": c.detail} for c in checks]})


async def daemon_list_endpoint(request) -> JSONResponse:
    """실행 중인 데몬 목록을 반환한다."""
    context: HttpContext = request.app.state.context
    return JSONResponse({"items": context.admin_service.daemon_list()})



"""MCP 심볼/호출자 도구(search_symbol/get_callers)를 제공한다."""

from __future__ import annotations

from sari.core.models import ErrorResponseDTO
from sari.db.repositories.lsp_tool_data_repository import LspToolDataRepository
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.mcp.tools.arg_parser import parse_non_empty_string, parse_optional_string, parse_positive_int
from sari.mcp.tools.admin_tools import validate_repo_argument
from sari.mcp.tools.pack1 import Pack1MetaDTO, pack1_error, pack1_success


class SearchSymbolTool:
    """search_symbol MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, lsp_repo: LspToolDataRepository) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._lsp_repo = lsp_repo

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """심볼 인덱스 검색 결과를 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)

        query_raw, query_error = parse_non_empty_string(arguments=arguments, key="query")
        if query_error is not None:
            return pack1_error(query_error)
        limit_raw, limit_error = parse_positive_int(arguments=arguments, key="limit", default=20)
        if limit_error is not None:
            return pack1_error(limit_error)
        path_prefix = parse_optional_string(arguments=arguments, key="path_prefix")

        repo = str(arguments["repo"])
        rows = self._lsp_repo.search_symbols(repo_root=repo, query=query_raw.strip(), limit=limit_raw, path_prefix=path_prefix)
        return pack1_success(
            {
                "items": [item.to_dict() for item in rows],
                "meta": Pack1MetaDTO(
                    candidate_count=len(rows),
                    resolved_count=len(rows),
                    cache_hit=True,
                    errors=[],
                ).to_dict(),
            }
        )


class GetCallersTool:
    """get_callers MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, lsp_repo: LspToolDataRepository) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._lsp_repo = lsp_repo

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """호출자 관계 조회 결과를 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)

        symbol_name = self._resolve_symbol_name(arguments)
        if symbol_name is None:
            return pack1_error(
                ErrorResponseDTO(code="ERR_SYMBOL_REQUIRED", message="symbol or symbol_id is required")
            )

        limit_raw, limit_error = parse_positive_int(arguments=arguments, key="limit", default=50)
        if limit_error is not None:
            return pack1_error(limit_error)

        repo = str(arguments["repo"])
        rows = self._lsp_repo.find_callers(repo_root=repo, symbol_name=symbol_name, limit=limit_raw)
        return pack1_success(
            {
                "items": [item.to_dict() for item in rows],
                "meta": Pack1MetaDTO(
                    candidate_count=len(rows),
                    resolved_count=len(rows),
                    cache_hit=True,
                    errors=[],
                ).to_dict(),
            }
        )

    def _resolve_symbol_name(self, arguments: dict[str, object]) -> str | None:
        """symbol/symbol_id 입력에서 검색 키를 결정한다."""
        symbol_raw = arguments.get("symbol")
        if isinstance(symbol_raw, str) and symbol_raw.strip() != "":
            return symbol_raw.strip()
        symbol_id_raw = arguments.get("symbol_id")
        if isinstance(symbol_id_raw, str) and symbol_id_raw.strip() != "":
            return symbol_id_raw.strip()
        return None



"""로컬 벡터 임베딩 기반 재정렬을 구현한다."""

from __future__ import annotations

import hashlib
import math
import re
from dataclasses import dataclass

from sari.core.models import SearchItemDTO, now_iso8601_utc
from sari.db.repositories.vector_embedding_repository import VectorEmbeddingRepository


@dataclass(frozen=True)
class VectorConfigDTO:
    """벡터 재정렬 설정을 표현한다."""

    enabled: bool = False
    model_id: str = "hashbow-v1"
    dim: int = 128
    candidate_k: int = 50
    rerank_k: int = 20
    blend_weight: float = 0.2
    min_similarity_threshold: float = 0.15
    max_vector_boost: float = 0.2
    min_token_count_for_rerank: int = 2
    apply_to_item_types: tuple[str, ...] = ("symbol", "file")


@dataclass(frozen=True)
class VectorRerankStatsDTO:
    """벡터 재정렬 수행 통계를 표현한다."""

    applied_count: int = 0
    skipped_count: int = 0


class VectorIndexSink:
    """파일 임베딩 생성/갱신 인터페이스를 제공한다."""

    def __init__(self, repository: VectorEmbeddingRepository, config: VectorConfigDTO) -> None:
        """저장소와 벡터 설정을 주입한다."""
        self._repository = repository
        self._config = config

    def upsert_file_embedding(self, repo_root: str, relative_path: str, content_hash: str, content_text: str) -> None:
        """파일 본문으로 임베딩을 생성해 저장한다."""
        if not self._config.enabled:
            return
        vector = _embed_text(content_text, self._config.dim)
        self._repository.upsert_file_embedding(
            repo_root=repo_root,
            relative_path=relative_path,
            content_hash=content_hash,
            model_id=self._config.model_id,
            dim=self._config.dim,
            vector=vector,
            updated_at=now_iso8601_utc(),
        )


class VectorReranker:
    """질의/파일 임베딩 유사도로 검색 결과를 재정렬한다."""

    def __init__(self, repository: VectorEmbeddingRepository, config: VectorConfigDTO) -> None:
        """저장소와 벡터 설정을 주입한다."""
        self._repository = repository
        self._config = config
        self._last_stats = VectorRerankStatsDTO()

    @property
    def config(self) -> VectorConfigDTO:
        """현재 벡터 설정을 반환한다."""
        return self._config

    @property
    def last_stats(self) -> VectorRerankStatsDTO:
        """직전 벡터 재정렬 실행 통계를 반환한다."""
        return self._last_stats

    def rerank(self, items: list[SearchItemDTO], query: str, limit: int) -> list[SearchItemDTO]:
        """상위 후보에 벡터 신호를 반영하되 최종 점수 계산은 수행하지 않는다."""
        if not self._config.enabled:
            self._last_stats = VectorRerankStatsDTO()
            return items[:limit]
        query_tokens = re.findall(r"[a-zA-Z0-9_]+", query.lower())
        if len(query_tokens) < max(1, self._config.min_token_count_for_rerank):
            self._last_stats = VectorRerankStatsDTO(applied_count=0, skipped_count=min(len(items), max(1, self._config.candidate_k)))
            return items[:limit]
        query_vector = self._get_or_create_query_embedding(query)
        if query_vector is None:
            self._last_stats = VectorRerankStatsDTO(applied_count=0, skipped_count=min(len(items), max(1, self._config.candidate_k)))
            return items[:limit]

        candidate_window = max(1, min(self._config.candidate_k, self._config.rerank_k))
        scoring_items: list[SearchItemDTO] = []
        applied_count = 0
        skipped_count = 0
        for item in items[:candidate_window]:
            if item.item_type not in self._config.apply_to_item_types:
                scoring_items.append(item)
                skipped_count += 1
                continue
            if item.content_hash is None:
                scoring_items.append(item)
                skipped_count += 1
                continue
            file_vector = self._repository.get_file_embedding(
                repo_root=item.repo,
                relative_path=item.relative_path,
                content_hash=item.content_hash,
                model_id=self._config.model_id,
            )
            if file_vector is None:
                scoring_items.append(item)
                skipped_count += 1
                continue
            similarity = _cosine_similarity(query_vector, file_vector)
            if similarity < self._config.min_similarity_threshold:
                scoring_items.append(
                    SearchItemDTO(
                        item_type=item.item_type,
                        repo=item.repo,
                        relative_path=item.relative_path,
                        score=item.score,
                        source=item.source,
                        name=item.name,
                        kind=item.kind,
                        content_hash=item.content_hash,
                        rrf_score=item.rrf_score,
                        importance_score=item.importance_score,
                        base_rrf_score=item.base_rrf_score,
                        importance_norm_score=item.importance_norm_score,
                        vector_norm_score=item.vector_norm_score,
                        hierarchy_score=item.hierarchy_score,
                        hierarchy_norm_score=item.hierarchy_norm_score,
                        symbol_key=item.symbol_key,
                        parent_symbol_key=item.parent_symbol_key,
                        depth=item.depth,
                        container_name=item.container_name,
                        ranking_components=item.ranking_components,
                        vector_score=similarity,
                        blended_score=item.blended_score,
                        final_score=item.final_score,
                    )
                )
                skipped_count += 1
                continue
            scoring_items.append(
                SearchItemDTO(
                    item_type=item.item_type,
                    repo=item.repo,
                    relative_path=item.relative_path,
                    score=item.score,
                    source=item.source,
                    name=item.name,
                    kind=item.kind,
                    content_hash=item.content_hash,
                    rrf_score=item.rrf_score,
                    importance_score=item.importance_score,
                    base_rrf_score=item.base_rrf_score,
                    importance_norm_score=item.importance_norm_score,
                    vector_norm_score=item.vector_norm_score,
                    hierarchy_score=item.hierarchy_score,
                    hierarchy_norm_score=item.hierarchy_norm_score,
                    symbol_key=item.symbol_key,
                    parent_symbol_key=item.parent_symbol_key,
                    depth=item.depth,
                    container_name=item.container_name,
                    ranking_components=item.ranking_components,
                    vector_score=similarity,
                    blended_score=item.blended_score,
                    final_score=item.final_score,
                )
            )
            applied_count += 1

        remain = items[candidate_window:]
        merged = [*scoring_items, *remain]
        self._last_stats = VectorRerankStatsDTO(applied_count=applied_count, skipped_count=skipped_count)
        return merged[:limit]

    def _get_or_create_query_embedding(self, query: str) -> list[float] | None:
        """질의 임베딩을 캐시 조회 후 없으면 생성한다."""
        normalized = query.strip().lower()
        if normalized == "":
            return None
        query_hash = hashlib.sha256(normalized.encode("utf-8")).hexdigest()
        cached = self._repository.get_query_embedding(query_hash=query_hash, model_id=self._config.model_id)
        if cached is not None:
            return cached
        vector = _embed_text(normalized, self._config.dim)
        self._repository.upsert_query_embedding(
            query_hash=query_hash,
            model_id=self._config.model_id,
            dim=self._config.dim,
            vector=vector,
            updated_at=now_iso8601_utc(),
        )
        return vector


def _embed_text(text: str, dim: int) -> list[float]:
    """텍스트를 해시 기반 bag-of-words 벡터로 변환한다."""
    vector = [0.0] * dim
    tokens = re.findall(r"[a-zA-Z0-9_]+", text.lower())
    for token in tokens:
        digest = hashlib.sha256(token.encode("utf-8")).digest()
        index = int.from_bytes(digest[:4], byteorder="big") % dim
        vector[index] += 1.0
    norm = math.sqrt(sum(value * value for value in vector))
    if norm <= 0.0:
        return vector
    return [value / norm for value in vector]


def _cosine_similarity(left: list[float], right: list[float]) -> float:
    """두 벡터의 코사인 유사도를 계산한다."""
    if len(left) == 0 or len(left) != len(right):
        return 0.0
    return float(sum(a * b for a, b in zip(left, right)))



"""
This module contains constants used in the LSP protocol.
"""


class LSPConstants:
    """
    This class contains constants used in the LSP protocol.
    """

    # the key for uri used to represent paths
    URI = "uri"

    # the key for range, which is a from and to position within a text document
    RANGE = "range"

    # A key used in LocationLink type, used as the span of the origin link
    ORIGIN_SELECTION_RANGE = "originSelectionRange"

    # A key used in LocationLink type, used as the target uri of the link
    TARGET_URI = "targetUri"

    # A key used in LocationLink type, used as the target range of the link
    TARGET_RANGE = "targetRange"

    # A key used in LocationLink type, used as the target selection range of the link
    TARGET_SELECTION_RANGE = "targetSelectionRange"

    # key for the textDocument field in the request
    TEXT_DOCUMENT = "textDocument"

    # key used to represent the language a document is in - "java", "csharp", etc.
    LANGUAGE_ID = "languageId"

    # key used to represent the version of a document (a shared value between the client and server)
    VERSION = "version"

    # key used to represent the text of a document being sent from the client to the server on open
    TEXT = "text"

    # key used to represent a position (line and colnum) within a text document
    POSITION = "position"

    # key used to represent the line number of a position
    LINE = "line"

    # key used to represent the column number of a position
    CHARACTER = "character"

    # key used to represent the changes made to a document
    CONTENT_CHANGES = "contentChanges"

    # key used to represent name of symbols
    NAME = "name"

    # key used to represent the kind of symbols
    KIND = "kind"

    # key used to represent children in document symbols
    CHILDREN = "children"

    # key used to represent the location in symbols
    LOCATION = "location"

    # Severity level of the diagnostic
    SEVERITY = "severity"

    # The message of the diagnostic
    MESSAGE = "message"



"""파이프라인 운영 제어 서비스를 구현한다."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone

from sari.core.exceptions import ErrorContext, ValidationError
from sari.core.models import (
    DeadJobActionResultDTO,
    DeadJobItemDTO,
    PipelineAlertSnapshotDTO,
    PipelineAutoControlStateDTO,
    PipelinePolicyDTO,
    now_iso8601_utc,
)
from sari.db.repositories.pipeline_control_state_repository import PipelineControlStateRepository
from sari.db.repositories.file_enrich_queue_repository import FileEnrichQueueRepository
from sari.db.repositories.pipeline_job_event_repository import PipelineJobEventRepository
from sari.db.repositories.pipeline_policy_repository import PipelinePolicyRepository


class PipelineControlService:
    """운영 정책/알람/DEAD 작업 제어를 담당한다."""

    def __init__(
        self,
        policy_repo: PipelinePolicyRepository,
        event_repo: PipelineJobEventRepository,
        queue_repo: FileEnrichQueueRepository,
        control_state_repo: PipelineControlStateRepository,
    ) -> None:
        """서비스 의존성을 저장한다."""
        self._policy_repo = policy_repo
        self._event_repo = event_repo
        self._queue_repo = queue_repo
        self._control_state_repo = control_state_repo

    def get_policy(self) -> PipelinePolicyDTO:
        """운영 정책을 조회한다."""
        return self._policy_repo.get_policy()

    def update_policy(
        self,
        deletion_hold: bool | None = None,
        l3_p95_threshold_ms: int | None = None,
        dead_ratio_threshold_bps: int | None = None,
        enrich_worker_count: int | None = None,
        bootstrap_mode_enabled: bool | None = None,
        bootstrap_l3_worker_count: int | None = None,
        bootstrap_l3_queue_max: int | None = None,
        bootstrap_exit_min_l2_coverage_bps: int | None = None,
        bootstrap_exit_max_sec: int | None = None,
        alert_window_sec: int | None = None,
    ) -> PipelinePolicyDTO:
        """운영 정책을 갱신한다."""
        if l3_p95_threshold_ms is not None and l3_p95_threshold_ms <= 0:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="l3_p95_threshold_ms는 1 이상이어야 합니다"))
        if dead_ratio_threshold_bps is not None and dead_ratio_threshold_bps <= 0:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="dead_ratio_threshold_bps는 1 이상이어야 합니다"))
        if enrich_worker_count is not None and enrich_worker_count <= 0:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="enrich_worker_count는 1 이상이어야 합니다"))
        if bootstrap_l3_worker_count is not None and bootstrap_l3_worker_count <= 0:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="bootstrap_l3_worker_count는 1 이상이어야 합니다"))
        if bootstrap_l3_queue_max is not None and bootstrap_l3_queue_max <= 0:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="bootstrap_l3_queue_max는 1 이상이어야 합니다"))
        if bootstrap_exit_min_l2_coverage_bps is not None and (
            bootstrap_exit_min_l2_coverage_bps <= 0 or bootstrap_exit_min_l2_coverage_bps > 10000
        ):
            raise ValidationError(
                ErrorContext(code="ERR_POLICY_INVALID", message="bootstrap_exit_min_l2_coverage_bps는 1..10000 범위여야 합니다")
            )
        if bootstrap_exit_max_sec is not None and bootstrap_exit_max_sec < 60:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="bootstrap_exit_max_sec는 60 이상이어야 합니다"))
        if alert_window_sec is not None and alert_window_sec < 60:
            raise ValidationError(ErrorContext(code="ERR_POLICY_INVALID", message="alert_window_sec는 60 이상이어야 합니다"))
        return self._policy_repo.update_policy(
            deletion_hold=deletion_hold,
            l3_p95_threshold_ms=l3_p95_threshold_ms,
            dead_ratio_threshold_bps=dead_ratio_threshold_bps,
            enrich_worker_count=enrich_worker_count,
            bootstrap_mode_enabled=bootstrap_mode_enabled,
            bootstrap_l3_worker_count=bootstrap_l3_worker_count,
            bootstrap_l3_queue_max=bootstrap_l3_queue_max,
            bootstrap_exit_min_l2_coverage_bps=bootstrap_exit_min_l2_coverage_bps,
            bootstrap_exit_max_sec=bootstrap_exit_max_sec,
            alert_window_sec=alert_window_sec,
        )

    def get_alert_status(self, now_iso: str | None = None) -> PipelineAlertSnapshotDTO:
        """최근 윈도우 기준 알람 스냅샷을 계산한다."""
        policy = self._policy_repo.get_policy()
        window_sec = self._policy_repo.get_alert_window_sec()
        now_dt = _parse_iso(now_iso) if now_iso is not None else datetime.now(timezone.utc)
        from_dt = now_dt - timedelta(seconds=window_sec)
        events = self._event_repo.list_window_events(from_iso=from_dt.isoformat())
        latencies: list[int] = []
        dead_count = 0
        for event in events:
            latency = int(event.get("latency_ms", 0))
            latencies.append(latency)
            if str(event.get("status", "")) == "DEAD":
                dead_count += 1
        event_count = len(events)
        l3_p95_ms = _percentile_95(latencies)
        dead_ratio_bps = 0
        if event_count > 0:
            dead_ratio_bps = int((dead_count * 10_000) / event_count)

        state = "ok"
        if dead_ratio_bps >= policy.dead_ratio_threshold_bps or l3_p95_ms >= policy.l3_p95_threshold_ms:
            state = "warn"
        if dead_ratio_bps >= policy.dead_ratio_threshold_bps * 2 or l3_p95_ms >= policy.l3_p95_threshold_ms * 2:
            state = "critical"

        return PipelineAlertSnapshotDTO(
            state=state,
            window_seconds=window_sec,
            event_count=event_count,
            dead_count=dead_count,
            dead_ratio_bps=dead_ratio_bps,
            l3_p95_ms=l3_p95_ms,
            threshold_dead_ratio_bps=policy.dead_ratio_threshold_bps,
            threshold_l3_p95_ms=policy.l3_p95_threshold_ms,
            policy=policy,
        )

    def list_dead_jobs(self, repo_root: str, limit: int) -> list[DeadJobItemDTO]:
        """DEAD 작업 목록을 조회한다."""
        if limit <= 0:
            raise ValidationError(ErrorContext(code="ERR_INVALID_LIMIT", message="limit는 1 이상이어야 합니다"))
        return self._queue_repo.list_dead(repo_root=repo_root, limit=limit)

    def requeue_dead_jobs(
        self,
        repo_root: str,
        limit: int,
        now_iso: str | None = None,
        all_scopes: bool = False,
    ) -> DeadJobActionResultDTO:
        """DEAD 작업을 재큐잉한다."""
        if limit <= 0:
            raise ValidationError(ErrorContext(code="ERR_INVALID_LIMIT", message="limit는 1 이상이어야 합니다"))
        applied_now = now_iso if now_iso is not None else now_iso8601_utc()
        if all_scopes:
            count = self._queue_repo.requeue_dead_all(now_iso=applied_now)
        else:
            count = self._queue_repo.requeue_dead(repo_root=repo_root, limit=limit, now_iso=applied_now)
        return DeadJobActionResultDTO(requeued_count=count, purged_count=0)

    def purge_dead_jobs(self, repo_root: str, limit: int, all_scopes: bool = False) -> DeadJobActionResultDTO:
        """DEAD 작업을 삭제한다."""
        if limit <= 0:
            raise ValidationError(ErrorContext(code="ERR_INVALID_LIMIT", message="limit는 1 이상이어야 합니다"))
        if all_scopes:
            count = self._queue_repo.purge_dead_all()
        else:
            count = self._queue_repo.purge_dead(repo_root=repo_root, limit=limit)
        return DeadJobActionResultDTO(requeued_count=0, purged_count=count)

    def get_auto_control_state(self) -> PipelineAutoControlStateDTO:
        """자동제어 상태를 조회한다."""
        return self._control_state_repo.get_state()

    def set_auto_hold_enabled(self, enabled: bool) -> PipelineAutoControlStateDTO:
        """자동제어 활성화를 설정한다."""
        action = "auto_enabled" if enabled else "auto_disabled"
        return self._control_state_repo.update_state(auto_hold_enabled=enabled, last_action=action)

    def evaluate_auto_hold(self, now_iso: str | None = None) -> dict[str, object]:
        """알람 상태를 기반으로 hold 자동 제어를 수행한다."""
        state = self._control_state_repo.get_state()
        if not state.auto_hold_enabled:
            return {"action": "AUTO_DISABLED", "changed": False, "auto_control": state.to_dict()}

        policy = self.get_policy()
        alert = self.get_alert_status(now_iso=now_iso)

        if alert.state in {"warn", "critical"} and not policy.deletion_hold:
            self.update_policy(deletion_hold=True)
            updated = self._control_state_repo.update_state(
                auto_hold_active=True,
                last_action=f"hold_enabled_by_auto:{alert.state}",
            )
            return {"action": "HOLD_ENABLED", "changed": True, "auto_control": updated.to_dict(), "alert": alert.to_dict()}

        if alert.state == "ok" and state.auto_hold_active and policy.deletion_hold:
            self.update_policy(deletion_hold=False)
            updated = self._control_state_repo.update_state(
                auto_hold_active=False,
                last_action="hold_released_by_auto",
            )
            return {"action": "HOLD_RELEASED", "changed": True, "auto_control": updated.to_dict(), "alert": alert.to_dict()}

        updated = self._control_state_repo.update_state(last_action=f"no_change:{alert.state}")
        return {"action": "NO_CHANGE", "changed": False, "auto_control": updated.to_dict(), "alert": alert.to_dict()}


def _parse_iso(raw: str) -> datetime:
    """ISO8601 문자열을 UTC datetime으로 변환한다."""
    parsed = datetime.fromisoformat(raw)
    if parsed.tzinfo is None:
        return parsed.replace(tzinfo=timezone.utc)
    return parsed.astimezone(timezone.utc)


def _percentile_95(values: list[int]) -> int:
    """정수 리스트의 95퍼센타일 값을 계산한다."""
    if len(values) == 0:
        return 0
    ordered = sorted(values)
    index = (len(ordered) * 95 + 99) // 100 - 1
    if index < 0:
        index = 0
    if index >= len(ordered):
        index = len(ordered) - 1
    return int(ordered[index])



"""stdio <-> daemon MCP framed 프록시를 구현한다."""

from __future__ import annotations

import os
import sys
from pathlib import Path

from sari.mcp.daemon_forward_policy import (
    StartDaemonFn,
    build_forward_error_message,
    default_start_daemon,
    forward_with_retry,
    resolve_target,
)
from sari.mcp.contracts import McpError, McpResponse
from sari.mcp.server_daemon_forward import DaemonForwardError, forward_once
from sari.mcp.tool_visibility import filter_tools_list_response_payload, is_hidden_tool_name
from sari.mcp.transport import MCP_MODE_FRAMED, McpTransport, McpTransportParseError


def _is_initialize_request(payload: dict[str, object]) -> bool:
    """payload가 initialize 요청인지 반환한다."""
    return str(payload.get("method", "")).strip() == "initialize"


def _is_tools_list_request(payload: dict[str, object]) -> bool:
    """payload가 tools/list 요청인지 반환한다."""
    return str(payload.get("method", "")).strip() == "tools/list"


def _extract_tools_call_name(payload: dict[str, object]) -> str | None:
    """tools/call 요청에서 도구명을 추출한다."""
    if str(payload.get("method", "")).strip() != "tools/call":
        return None
    params = payload.get("params")
    if not isinstance(params, dict):
        return None
    name = params.get("name")
    if not isinstance(name, str):
        return None
    normalized = name.strip()
    if normalized == "":
        return None
    return normalized


def _is_draining_response(payload: dict[str, object]) -> bool:
    """daemon 응답이 draining 오류인지 판정한다."""
    error_obj = payload.get("error")
    if not isinstance(error_obj, dict):
        return False
    code_raw = error_obj.get("code")
    message_raw = str(error_obj.get("message", "")).lower()
    if isinstance(code_raw, int) and code_raw == -32001 and "draining" in message_raw:
        return True
    return "draining" in message_raw


def _is_draining_reconnect_enabled() -> bool:
    """draining 자동 재연결 활성화 여부를 반환한다."""
    raw = os.getenv("SARI_MCP_DRAINING_RECONNECT", "1").strip().lower()
    return raw in {"1", "true", "yes", "on"}


def _retry_after_draining(
    *,
    payload: dict[str, object],
    db_path: Path,
    workspace_root: str | None,
    host_override: str | None,
    port_override: int | None,
    timeout_sec: float,
    initialize_payload: dict[str, object] | None,
) -> dict[str, object]:
    """draining 응답 이후 endpoint 재해석 + initialize 재전송 + 원요청 재시도한다."""
    host_retry, port_retry = resolve_target(db_path, workspace_root, host_override, port_override)
    if initialize_payload is not None and not _is_initialize_request(payload):
        initialize_response = forward_once(initialize_payload, host_retry, port_retry, timeout_sec)
        if _is_draining_response(initialize_response):
            raise ValueError("ERR_DAEMON_DRAINING_RECONNECT_FAILED: initialize replay still draining")
    retried = forward_once(payload, host_retry, port_retry, timeout_sec)
    if _is_draining_response(retried):
        raise ValueError("ERR_DAEMON_DRAINING_RECONNECT_FAILED: request replay still draining")
    return retried


def run_stdio_proxy(
    db_path: Path,
    workspace_root: str | None = None,
    host_override: str | None = None,
    port_override: int | None = None,
    timeout_sec: float = 2.0,
    auto_start_on_failure: bool = True,
    start_daemon_fn: StartDaemonFn | None = None,
) -> int:
    """표준 입출력 MCP 요청을 daemon endpoint로 중계한다."""
    input_stream = getattr(sys.stdin, "buffer", sys.stdin)
    output_stream = getattr(sys.stdout, "buffer", sys.stdout)
    transport = McpTransport(input_stream=input_stream, output_stream=output_stream, allow_jsonl=True)
    transport.default_mode = MCP_MODE_FRAMED
    daemon_starter = default_start_daemon if start_daemon_fn is None else start_daemon_fn
    initialize_payload: dict[str, object] | None = None

    while True:
        try:
            read_result = transport.read_message()
        except McpTransportParseError as exc:
            parse_response = McpResponse(request_id=None, result=None, error=McpError(code=-32700, message=str(exc)))
            transport.write_message(parse_response.to_dict(), mode=exc.mode)
            continue

        if read_result is None:
            return 0
        payload, mode = read_result
        request_id = payload.get("id")
        is_notification = request_id is None
        hidden_tool_name = _extract_tools_call_name(payload)
        if hidden_tool_name is not None and is_hidden_tool_name(hidden_tool_name):
            if is_notification:
                continue
            response = McpResponse(
                request_id=request_id,
                result=None,
                error=McpError(code=-32601, message="tool not found"),
            )
            transport.write_message(response.to_dict(), mode=mode)
            continue
        if _is_initialize_request(payload):
            initialize_payload = payload
        try:
            forwarded = forward_with_retry(
                request=payload,
                db_path=db_path,
                workspace_root=workspace_root,
                host_override=host_override,
                port_override=port_override,
                timeout_sec=timeout_sec,
                auto_start_on_failure=auto_start_on_failure,
                start_daemon_fn=daemon_starter,
                resolve_target_fn=resolve_target,
                forward_once_fn=forward_once,
            )
            if _is_draining_reconnect_enabled() and _is_draining_response(forwarded):
                forwarded = _retry_after_draining(
                    payload=payload,
                    db_path=db_path,
                    workspace_root=workspace_root,
                    host_override=host_override,
                    port_override=port_override,
                    timeout_sec=timeout_sec,
                    initialize_payload=initialize_payload,
                )
            if _is_tools_list_request(payload):
                forwarded = filter_tools_list_response_payload(forwarded)
            if is_notification:
                continue
            transport.write_message(forwarded, mode=mode)
        except (DaemonForwardError, OSError, TimeoutError, ValueError) as exc:
            if is_notification:
                continue
            response = McpResponse(
                request_id=request_id,
                result=None,
                error=McpError(code=-32002, message=build_forward_error_message(exc)),
            )
            transport.write_message(response.to_dict(), mode=mode)



from __future__ import annotations

from enum import Enum, IntEnum, IntFlag
from typing import Literal, NotRequired, Union

from typing_extensions import TypedDict


URI = str
DocumentUri = str
Uint = int
RegExp = str


class SemanticTokenTypes(Enum):
    """A set of predefined token types. This set is not fixed
    an clients can specify additional token types via the
    corresponding client capabilities.

    @since 3.16.0
    """

    Namespace = "namespace"
    Type = "type"
    """ Represents a generic type. Acts as a fallback for types which can't be mapped to
    a specific type like class or enum. """
    Class = "class"
    Enum = "enum"
    Interface = "interface"
    Struct = "struct"
    TypeParameter = "typeParameter"
    Parameter = "parameter"
    Variable = "variable"
    Property = "property"
    EnumMember = "enumMember"
    Event = "event"
    Function = "function"
    Method = "method"
    Macro = "macro"
    Keyword = "keyword"
    Modifier = "modifier"
    Comment = "comment"
    String = "string"
    Number = "number"
    Regexp = "regexp"
    Operator = "operator"
    Decorator = "decorator"
    """ @since 3.17.0 """


class SemanticTokenModifiers(Enum):
    """A set of predefined token modifiers. This set is not fixed
    an clients can specify additional token types via the
    corresponding client capabilities.

    @since 3.16.0
    """

    Declaration = "declaration"
    Definition = "definition"
    Readonly = "readonly"
    Static = "static"
    Deprecated = "deprecated"
    Abstract = "abstract"
    Async = "async"
    Modification = "modification"
    Documentation = "documentation"
    DefaultLibrary = "defaultLibrary"


class DocumentDiagnosticReportKind(Enum):
    """The document diagnostic report kinds.

    @since 3.17.0
    """

    Full = "full"
    """ A diagnostic report with a full
    set of problems. """
    Unchanged = "unchanged"
    """ A report indicating that the last
    returned report is still accurate. """


class ErrorCodes(IntEnum):
    """Predefined error codes."""

    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    ServerNotInitialized = -32002
    """ Error code indicating that a server received a notification or
    request before the server has received the `initialize` request. """
    UnknownErrorCode = -32001


class LSPErrorCodes(IntEnum):
    RequestFailed = -32803
    """ A request failed but it was syntactically correct, e.g the
    method name was known and the parameters were valid. The error
    message should contain human readable information about why
    the request failed.

    @since 3.17.0 """
    ServerCancelled = -32802
    """ The server cancelled the request. This error code should
    only be used for requests that explicitly support being
    server cancellable.

    @since 3.17.0 """
    ContentModified = -32801
    """ The server detected that the content of a document got
    modified outside normal conditions. A server should
    NOT send this error code if it detects a content change
    in it unprocessed messages. The result even computed
    on an older state might still be useful for the client.

    If a client decides that a result is not of any use anymore
    the client should cancel the request. """
    RequestCancelled = -32800
    """ The client has canceled a request and a server as detected
    the cancel. """


class FoldingRangeKind(Enum):
    """A set of predefined range kinds."""

    Comment = "comment"
    """ Folding range for a comment """
    Imports = "imports"
    """ Folding range for an import or include """
    Region = "region"
    """ Folding range for a region (e.g. `#region`) """


class SymbolKind(IntEnum):
    """A symbol kind."""

    File = 1
    Module = 2
    Namespace = 3
    Package = 4
    """
    Represents a package or simply a directory in the filesystem
    """
    Class = 5
    Method = 6
    Property = 7
    Field = 8
    Constructor = 9
    Enum = 10
    Interface = 11
    Function = 12
    Variable = 13
    Constant = 14
    String = 15
    Number = 16
    Boolean = 17
    Array = 18
    Object = 19
    Key = 20
    Null = 21
    EnumMember = 22
    Struct = 23
    Event = 24
    Operator = 25
    TypeParameter = 26

    @classmethod
    def from_int(cls, value: int) -> "SymbolKind":
        for symbol_kind in cls:
            if symbol_kind.value == value:
                return symbol_kind
        raise ValueError(f"Invalid symbol kind: {value}")


class SymbolTag(IntEnum):
    """Symbol tags are extra annotations that tweak the rendering of a symbol.

    @since 3.16
    """

    Deprecated = 1
    """ Render a symbol as obsolete, usually using a strike-out. """


class UniquenessLevel(Enum):
    """Moniker uniqueness level to define scope of the moniker.

    @since 3.16.0
    """

    Document = "document"
    """ The moniker is only unique inside a document """
    Project = "project"
    """ The moniker is unique inside a project for which a dump got created """
    Group = "group"
    """ The moniker is unique inside the group to which a project belongs """
    Scheme = "scheme"
    """ The moniker is unique inside the moniker scheme. """
    Global = "global"
    """ The moniker is globally unique """


class MonikerKind(Enum):
    """The moniker kind.

    @since 3.16.0
    """

    Import = "import"
    """ The moniker represent a symbol that is imported into a project """
    Export = "export"
    """ The moniker represents a symbol that is exported from a project """
    Local = "local"
    """ The moniker represents a symbol that is local to a project (e.g. a local
    variable of a function, a class not visible outside the project, ...) """


class InlayHintKind(IntEnum):
    """Inlay hint kinds.

    @since 3.17.0
    """

    Type = 1
    """ An inlay hint that for a type annotation. """
    Parameter = 2
    """ An inlay hint that is for a parameter. """


class MessageType(IntEnum):
    """The message type"""

    Error = 1
    """ An error message. """
    Warning = 2
    """ A warning message. """
    Info = 3
    """ An information message. """
    Log = 4
    """ A log message. """


class TextDocumentSyncKind(IntEnum):
    """Defines how the host (editor) should sync
    document changes to the language server.
    """

    None_ = 0
    """ Documents should not be synced at all. """
    Full = 1
    """ Documents are synced by always sending the full content
    of the document. """
    Incremental = 2
    """ Documents are synced by sending the full content on open.
    After that only incremental updates to the document are
    send. """


class TextDocumentSaveReason(IntEnum):
    """Represents reasons why a text document is saved."""

    Manual = 1
    """ Manually triggered, e.g. by the user pressing save, by starting debugging,
    or by an API call. """
    AfterDelay = 2
    """ Automatic after a delay. """
    FocusOut = 3
    """ When the editor lost focus. """


class CompletionItemKind(IntEnum):
    """The kind of a completion entry."""

    Text = 1
    Method = 2
    Function = 3
    Constructor = 4
    Field = 5
    Variable = 6
    Class = 7
    Interface = 8
    Module = 9
    Property = 10
    Unit = 11
    Value = 12
    Enum = 13
    Keyword = 14
    Snippet = 15
    Color = 16
    File = 17
    Reference = 18
    Folder = 19
    EnumMember = 20
    Constant = 21
    Struct = 22
    Event = 23
    Operator = 24
    TypeParameter = 25


class CompletionItemTag(IntEnum):
    """Completion item tags are extra annotations that tweak the rendering of a completion
    item.

    @since 3.15.0
    """

    Deprecated = 1
    """ Render a completion as obsolete, usually using a strike-out. """


class InsertTextFormat(IntEnum):
    """Defines whether the insert text in a completion item should be interpreted as
    plain text or a snippet.
    """

    PlainText = 1
    """ The primary text to be inserted is treated as a plain string. """
    Snippet = 2
    """ The primary text to be inserted is treated as a snippet.

    A snippet can define tab stops and placeholders with `$1`, `$2`
    and `${3:foo}`. `$0` defines the final tab stop, it defaults to
    the end of the snippet. Placeholders with equal identifiers are linked,
    that is typing in one will update others too.

    See also: https://microsoft.github.io/language-server-protocol/specifications/specification-current/#snippet_syntax """


class InsertTextMode(IntEnum):
    """How whitespace and indentation is handled during completion
    item insertion.

    @since 3.16.0
    """

    AsIs = 1
    """ The insertion or replace strings is taken as it is. If the
    value is multi line the lines below the cursor will be
    inserted using the indentation defined in the string value.
    The client will not apply any kind of adjustments to the
    string. """
    AdjustIndentation = 2
    """ The editor adjusts leading whitespace of new lines so that
    they match the indentation up to the cursor of the line for
    which the item is accepted.

    Consider a line like this: <2tabs><cursor><3tabs>foo. Accepting a
    multi line completion item is indented using 2 tabs and all
    following lines inserted will be indented using 2 tabs as well. """


class DocumentHighlightKind(IntEnum):
    """A document highlight kind."""

    Text = 1
    """ A textual occurrence. """
    Read = 2
    """ Read-access of a symbol, like reading a variable. """
    Write = 3
    """ Write-access of a symbol, like writing to a variable. """


class CodeActionKind(Enum):
    """A set of predefined code action kinds"""

    Empty = ""
    """ Empty kind. """
    QuickFix = "quickfix"
    """ Base kind for quickfix actions: 'quickfix' """
    Refactor = "refactor"
    """ Base kind for refactoring actions: 'refactor' """
    RefactorExtract = "refactor.extract"
    """ Base kind for refactoring extraction actions: 'refactor.extract'

    Example extract actions:

    - Extract method
    - Extract function
    - Extract variable
    - Extract interface from class
    - ... """
    RefactorInline = "refactor.inline"
    """ Base kind for refactoring inline actions: 'refactor.inline'

    Example inline actions:

    - Inline function
    - Inline variable
    - Inline constant
    - ... """
    RefactorRewrite = "refactor.rewrite"
    """ Base kind for refactoring rewrite actions: 'refactor.rewrite'

    Example rewrite actions:

    - Convert JavaScript function to class
    - Add or remove parameter
    - Encapsulate field
    - Make method static
    - Move method to base class
    - ... """
    Source = "source"
    """ Base kind for source actions: `source`

    Source code actions apply to the entire file. """
    SourceOrganizeImports = "source.organizeImports"
    """ Base kind for an organize imports source action: `source.organizeImports` """
    SourceFixAll = "source.fixAll"
    """ Base kind for auto-fix source actions: `source.fixAll`.

    Fix all actions automatically fix errors that have a clear fix that do not require user input.
    They should not suppress errors or perform unsafe fixes such as generating new types or classes.

    @since 3.15.0 """


class TraceValues(Enum):
    Off = "off"
    """ Turn tracing off. """
    Messages = "messages"
    """ Trace messages only. """
    Verbose = "verbose"
    """ Verbose message tracing. """


class MarkupKind(Enum):
    """Describes the content type that a client supports in various
    result literals like `Hover`, `ParameterInfo` or `CompletionItem`.

    Please note that `MarkupKinds` must not start with a `$`. This kinds
    are reserved for internal usage.
    """

    PlainText = "plaintext"
    """ Plain text is supported as a content format """
    Markdown = "markdown"
    """ Markdown is supported as a content format """


class PositionEncodingKind(Enum):
    """A set of predefined position encoding kinds.

    @since 3.17.0
    """

    UTF8 = "utf-8"
    """ Character offsets count UTF-8 code units. """
    UTF16 = "utf-16"
    """ Character offsets count UTF-16 code units.

    This is the default and must always be supported
    by servers """
    UTF32 = "utf-32"
    """ Character offsets count UTF-32 code units.

    Implementation note: these are the same as Unicode code points,
    so this `PositionEncodingKind` may also be used for an
    encoding-agnostic representation of character offsets. """


class FileChangeType(IntEnum):
    """The file event type"""

    Created = 1
    """ The file got created. """
    Changed = 2
    """ The file got changed. """
    Deleted = 3
    """ The file got deleted. """


class WatchKind(IntFlag):
    Create = 1
    """ Interested in create events. """
    Change = 2
    """ Interested in change events """
    Delete = 4
    """ Interested in delete events """


class DiagnosticSeverity(IntEnum):
    """The diagnostic's severity."""

    Error = 1
    """ Reports an error. """
    Warning = 2
    """ Reports a warning. """
    Information = 3
    """ Reports an information. """
    Hint = 4
    """ Reports a hint. """


class DiagnosticTag(IntEnum):
    """The diagnostic tags.

    @since 3.15.0
    """

    Unnecessary = 1
    """ Unused or unnecessary code.

    Clients are allowed to render diagnostics with this tag faded out instead of having
    an error squiggle. """
    Deprecated = 2
    """ Deprecated or obsolete code.

    Clients are allowed to rendered diagnostics with this tag strike through. """


class CompletionTriggerKind(IntEnum):
    """How a completion was triggered"""

    Invoked = 1
    """ Completion was triggered by typing an identifier (24x7 code
    complete), manual invocation (e.g Ctrl+Space) or via API. """
    TriggerCharacter = 2
    """ Completion was triggered by a trigger character specified by
    the `triggerCharacters` properties of the `CompletionRegistrationOptions`. """
    TriggerForIncompleteCompletions = 3
    """ Completion was re-triggered as current completion list is incomplete """


class SignatureHelpTriggerKind(IntEnum):
    """How a signature help was triggered.

    @since 3.15.0
    """

    Invoked = 1
    """ Signature help was invoked manually by the user or by a command. """
    TriggerCharacter = 2
    """ Signature help was triggered by a trigger character. """
    ContentChange = 3
    """ Signature help was triggered by the cursor moving or by the document content changing. """


class CodeActionTriggerKind(IntEnum):
    """The reason why code actions were requested.

    @since 3.17.0
    """

    Invoked = 1
    """ Code actions were explicitly requested by the user or by an extension. """
    Automatic = 2
    """ Code actions were requested automatically.

    This typically happens when current selection in a file changes, but can
    also be triggered when file content changes. """


class FileOperationPatternKind(Enum):
    """A pattern kind describing if a glob pattern matches a file a folder or
    both.

    @since 3.16.0
    """

    File = "file"
    """ The pattern matches a file only. """
    Folder = "folder"
    """ The pattern matches a folder only. """


class NotebookCellKind(IntEnum):
    """A notebook cell kind.

    @since 3.17.0
    """

    Markup = 1
    """ A markup-cell is formatted source that is used for display. """
    Code = 2
    """ A code-cell is source code. """


class ResourceOperationKind(Enum):
    Create = "create"
    """ Supports creating new files and folders. """
    Rename = "rename"
    """ Supports renaming existing files and folders. """
    Delete = "delete"
    """ Supports deleting existing files and folders. """


class FailureHandlingKind(Enum):
    Abort = "abort"
    """ Applying the workspace change is simply aborted if one of the changes provided
    fails. All operations executed before the failing operation stay executed. """
    Transactional = "transactional"
    """ All operations are executed transactional. That means they either all
    succeed or no changes at all are applied to the workspace. """
    TextOnlyTransactional = "textOnlyTransactional"
    """ If the workspace edit contains only textual file changes they are executed transactional.
    If resource changes (create, rename or delete file) are part of the change the failure
    handling strategy is abort. """
    Undo = "undo"
    """ The client tries to undo the operations already executed. But there is no
    guarantee that this is succeeding. """


class PrepareSupportDefaultBehavior(IntEnum):
    Identifier = 1
    """ The client's default behavior is to select the identifier
    according the to language's syntax rule. """


class TokenFormat(Enum):
    Relative = "relative"


Definition = Union["Location", list["Location"]]
""" The definition of a symbol represented as one or many {@link Location locations}.
For most programming languages there is only one location at which a symbol is
defined.

Servers should prefer returning `DefinitionLink` over `Definition` if supported
by the client. """

DefinitionLink = "LocationLink"
""" Information about where a symbol is defined.

Provides additional metadata over normal {@link Location location} definitions, including the range of
the defining symbol """

LSPArray = list["LSPAny"]
""" LSP arrays.
@since 3.17.0 """

LSPAny = Union["LSPObject", "LSPArray", str, int, Uint, float, bool, None]
""" The LSP any type.
Please note that strictly speaking a property with the value `undefined`
can't be converted into JSON preserving the property name. However for
convenience it is allowed and assumed that all these properties are
optional as well.
@since 3.17.0 """

Declaration = Union["Location", list["Location"]]
""" The declaration of a symbol representation as one or many {@link Location locations}. """

DeclarationLink = "LocationLink"
""" Information about where a symbol is declared.

Provides additional metadata over normal {@link Location location} declarations, including the range of
the declaring symbol.

Servers should prefer returning `DeclarationLink` over `Declaration` if supported
by the client. """

InlineValue = Union["InlineValueText", "InlineValueVariableLookup", "InlineValueEvaluatableExpression"]
""" Inline value information can be provided by different means:
- directly as a text value (class InlineValueText).
- as a name to use for a variable lookup (class InlineValueVariableLookup)
- as an evaluatable expression (class InlineValueEvaluatableExpression)
The InlineValue types combines all inline value types into one type.

@since 3.17.0 """

DocumentDiagnosticReport = Union["RelatedFullDocumentDiagnosticReport", "RelatedUnchangedDocumentDiagnosticReport"]
""" The result of a document diagnostic pull request. A report can
either be a full report containing all diagnostics for the
requested document or an unchanged report indicating that nothing
has changed in terms of diagnostics in comparison to the last
pull request.

@since 3.17.0 """

PrepareRenameResult = Union["Range", "__PrepareRenameResult_Type_1", "__PrepareRenameResult_Type_2"]

DocumentSelector = list["DocumentFilter"]
""" A document selector is the combination of one or many document filters.

@sample `let sel:DocumentSelector = [{ language: 'typescript' }, { language: 'json', pattern: '**/tsconfig.json' }]`;

The use of a string as a document filter is deprecated @since 3.16.0. """

ProgressToken = Union[int, str]

ChangeAnnotationIdentifier = str
""" An identifier to refer to a change annotation stored with a workspace edit. """

WorkspaceDocumentDiagnosticReport = Union[
    "WorkspaceFullDocumentDiagnosticReport",
    "WorkspaceUnchangedDocumentDiagnosticReport",
]
""" A workspace diagnostic document report.

@since 3.17.0 """

TextDocumentContentChangeEvent = Union["__TextDocumentContentChangeEvent_Type_1", "__TextDocumentContentChangeEvent_Type_2"]
""" An event describing a change to a text document. If only a text is provided
it is considered to be the full content of the document. """

MarkedString = Union[str, "__MarkedString_Type_1"]
""" MarkedString can be used to render human readable text. It is either a markdown string
or a code-block that provides a language and a code snippet. The language identifier
is semantically equal to the optional language identifier in fenced code blocks in GitHub
issues. See https://help.github.com/articles/creating-and-highlighting-code-blocks/#syntax-highlighting

The pair of a language and a value is an equivalent to markdown:
```${language}
${value}
```

Note that markdown strings will be sanitized - that means html will be escaped.
@deprecated use MarkupContent instead. """

DocumentFilter = Union["TextDocumentFilter", "NotebookCellTextDocumentFilter"]
""" A document filter describes a top level text document or
a notebook cell document.

@since 3.17.0 - proposed support for NotebookCellTextDocumentFilter. """

LSPObject = dict[str, "LSPAny"]
""" LSP object definition.
@since 3.17.0 """

GlobPattern = Union["Pattern", "RelativePattern"]
""" The glob pattern. Either a string pattern or a relative pattern.

@since 3.17.0 """

TextDocumentFilter = Union[
    "__TextDocumentFilter_Type_1",
    "__TextDocumentFilter_Type_2",
    "__TextDocumentFilter_Type_3",
]
""" A document filter denotes a document by different properties like
the {@link TextDocument.languageId language}, the {@link Uri.scheme scheme} of
its resource, or a glob-pattern that is applied to the {@link TextDocument.fileName path}.

Glob patterns can have the following syntax:
- `*` to match one or more characters in a path segment
- `?` to match on one character in a path segment
- `**` to match any number of path segments, including none
- `{}` to group sub patterns into an OR expression. (e.g. `**\u200b/*.{ts,js}` matches all TypeScript and JavaScript files)
- `[]` to declare a range of characters to match in a path segment (e.g., `example.[0-9]` to match on `example.0`, `example.1`, …)
- `[!...]` to negate a range of characters to match in a path segment (e.g., `example.[!0-9]` to match on `example.a`, `example.b`, but not `example.0`)

@sample A language filter that applies to typescript files on disk: `{ language: 'typescript', scheme: 'file' }`
@sample A language filter that applies to all package.json paths: `{ language: 'json', pattern: '**package.json' }`

@since 3.17.0 """

NotebookDocumentFilter = Union[
    "__NotebookDocumentFilter_Type_1",
    "__NotebookDocumentFilter_Type_2",
    "__NotebookDocumentFilter_Type_3",
]
""" A notebook document filter denotes a notebook document by
different properties. The properties will be match
against the notebook's URI (same as with documents)

@since 3.17.0 """



"""
Provides Python specific instantiation of the LanguageServer class. Contains various configurations and settings specific to Python.
"""

import logging
import os
import pathlib
import re
import sys
import threading
from typing import cast

from overrides import override

from solidlsp.ls import LanguageServerDependencyProvider, LanguageServerDependencyProviderSinglePath, SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)


class PyrightServer(SolidLanguageServer):
    """
    Provides Python specific instantiation of the LanguageServer class using Pyright.
    Contains various configurations and settings specific to Python.
    """

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a PyrightServer instance. This class is not meant to be instantiated directly.
        Use LanguageServer.create() instead.
        """
        super().__init__(
            config,
            repository_root_path,
            None,
            "python",
            solidlsp_settings,
        )

        # Event to signal when initial workspace analysis is complete
        self.analysis_complete = threading.Event()
        self.found_source_files = False

    def _create_dependency_provider(self) -> LanguageServerDependencyProvider:
        return self.DependencyProvider(self._custom_settings, self._ls_resources_dir)

    def _requires_open_file_for_document_symbols(self) -> bool:
        return False

    class DependencyProvider(LanguageServerDependencyProviderSinglePath):
        def _get_or_install_core_dependency(self) -> str:
            return sys.executable

        def _create_launch_command(self, core_path: str) -> list[str]:
            return [core_path, "-m", "pyright.langserver", "--stdio"]

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        return super().is_ignored_dirname(dirname) or dirname in ["venv", "__pycache__"]

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Pyright Language Server.
        """
        # Create basic initialization parameters
        initialize_params = {  # type: ignore
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": pathlib.Path(repository_absolute_path).as_uri(),
            "initializationOptions": {
                "exclude": [
                    "**/__pycache__",
                    "**/.venv",
                    "**/.env",
                    "**/build",
                    "**/dist",
                    "**/.pixi",
                ],
                "reportMissingImports": "error",
            },
            "capabilities": {
                "workspace": {
                    "workspaceEdit": {"documentChanges": True},
                    "didChangeConfiguration": {"dynamicRegistration": True},
                    "didChangeWatchedFiles": {"dynamicRegistration": True},
                    "symbol": {
                        "dynamicRegistration": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                    },
                    "executeCommand": {"dynamicRegistration": True},
                },
                "textDocument": {
                    "synchronization": {"dynamicRegistration": True, "willSave": True, "willSaveWaitUntil": True, "didSave": True},
                    "hover": {"dynamicRegistration": True, "contentFormat": ["markdown", "plaintext"]},
                    "signatureHelp": {
                        "dynamicRegistration": True,
                        "signatureInformation": {
                            "documentationFormat": ["markdown", "plaintext"],
                            "parameterInformation": {"labelOffsetSupport": True},
                        },
                    },
                    "definition": {"dynamicRegistration": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                        "hierarchicalDocumentSymbolSupport": True,
                    },
                    "publishDiagnostics": {"relatedInformation": True},
                },
            },
            "workspaceFolders": [
                {"uri": pathlib.Path(repository_absolute_path).as_uri(), "name": os.path.basename(repository_absolute_path)}
            ],
        }

        return cast(InitializeParams, initialize_params)

    def _start_server(self) -> None:
        """
        Starts the Pyright Language Server and waits for initial workspace analysis to complete.

        This prevents zombie processes by ensuring Pyright has finished its initial background
        tasks before we consider the server ready.

        Usage:
        ```
        async with lsp.start_server():
            # LanguageServer has been initialized and workspace analysis is complete
            await lsp.request_definition(...)
            await lsp.request_references(...)
            # Shutdown the LanguageServer on exit from scope
        # LanguageServer has been shutdown cleanly
        ```
        """

        def execute_client_command_handler(params: dict) -> list:
            return []

        def do_nothing(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            """
            Monitor Pyright's log messages to detect when initial analysis is complete.
            Pyright logs "Found X source files" when it finishes scanning the workspace.
            """
            message_text = msg.get("message", "")
            log.info(f"LSP: window/logMessage: {message_text}")

            # Look for "Found X source files" which indicates workspace scanning is complete
            # Unfortunately, pyright is unreliable and there seems to be no better way
            if re.search(r"Found \d+ source files?", message_text):
                log.info("Pyright workspace scanning complete")
                self.found_source_files = True
                self.analysis_complete.set()

        def check_experimental_status(params: dict) -> None:
            """
            Also listen for experimental/serverStatus as a backup signal
            """
            if params.get("quiescent") == True:
                log.info("Received experimental/serverStatus with quiescent=true")
                if not self.found_source_files:
                    self.analysis_complete.set()

        # Set up notification handlers
        self.server.on_request("client/registerCapability", do_nothing)
        self.server.on_notification("language/status", do_nothing)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_request("workspace/executeClientCommand", execute_client_command_handler)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)
        self.server.on_notification("language/actionableNotification", do_nothing)
        self.server.on_notification("experimental/serverStatus", check_experimental_status)

        log.info("Starting pyright-langserver server process")
        self.server.start()

        # Send proper initialization parameters
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to pyright server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)
        log.info(f"Received initialize response from pyright server: {init_response}")

        # Verify that the server supports our required features
        assert "textDocumentSync" in init_response["capabilities"]
        assert "completionProvider" in init_response["capabilities"]
        assert "definitionProvider" in init_response["capabilities"]

        # Complete the initialization handshake
        self.server.notify.initialized({})

        # Wait for Pyright to complete its initial workspace analysis
        # This prevents zombie processes by ensuring background tasks finish
        log.info("Waiting for Pyright to complete initial workspace analysis...")
        if self.analysis_complete.wait(timeout=0.5):
            log.info("Pyright initial analysis complete, server ready")
        else:
            log.warning("Timeout waiting for Pyright analysis completion, proceeding anyway")
            # Fallback: assume analysis is complete after timeout
            self.analysis_complete.set()



"""MCP daemon forward 공통 정책을 제공한다."""

from __future__ import annotations

from dataclasses import replace
from pathlib import Path
import socket
import time
from typing import Callable

from sari.core.config import AppConfig
from sari.core.daemon_resolver import resolve_daemon_address
from sari.core.exceptions import DaemonError
from sari.db.repositories.daemon_registry_repository import DaemonRegistryRepository
from sari.db.repositories.runtime_repository import RuntimeRepository
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.mcp.server_daemon_forward import DaemonForwardError
from sari.services.daemon_service import DaemonService

StartDaemonFn = Callable[[Path, str | None], bool]
ResolveTargetFn = Callable[[Path, str | None, str | None, int | None], tuple[str, int]]
ForwardOnceFn = Callable[[dict[str, object], str, int, float], dict[str, object]]

FORWARD_METHODS = frozenset({"tools/list", "tools/call"})
POST_START_RETRY_MAX = 8
POST_START_RETRY_WAIT_SEC = 0.15
INITIALIZE_POST_START_RETRY_MAX = 40
INITIALIZE_POST_START_RETRY_WAIT_SEC = 0.2
DAEMON_START_READY_TIMEOUT_SEC = 20.0
DAEMON_START_READY_POLL_SEC = 0.2


def should_forward_to_daemon(proxy_enabled: bool, method: str) -> bool:
    """daemon forward 대상 메서드 여부를 반환한다."""
    if not proxy_enabled:
        return False
    return method in FORWARD_METHODS


def extract_workspace_root(payload: dict[str, object]) -> str | None:
    """요청 payload에서 repo(workspace_root)를 추출한다."""
    params = payload.get("params")
    if not isinstance(params, dict):
        return None
    arguments = params.get("arguments")
    if not isinstance(arguments, dict):
        return None
    repo = arguments.get("repo")
    if not isinstance(repo, str):
        return None
    stripped = repo.strip()
    if stripped == "":
        return None
    return stripped


def resolve_target(
    db_path: Path,
    workspace_root: str | None,
    host_override: str | None,
    port_override: int | None,
) -> tuple[str, int]:
    """override 우선으로 daemon endpoint를 결정한다."""
    if host_override is not None and host_override.strip() != "" and port_override is not None:
        return host_override.strip(), port_override
    return resolve_daemon_address(db_path=db_path, workspace_root=workspace_root)


def forward_with_retry(
    request: dict[str, object],
    db_path: Path,
    workspace_root: str | None,
    host_override: str | None,
    port_override: int | None,
    timeout_sec: float,
    auto_start_on_failure: bool,
    start_daemon_fn: StartDaemonFn,
    resolve_target_fn: ResolveTargetFn,
    forward_once_fn: ForwardOnceFn,
) -> dict[str, object]:
    """forward 실패 시 자동 기동 후 1회 재시도한다."""
    retry_max, retry_wait_sec = resolve_post_start_retry_policy(request)
    host, port = resolve_target_fn(db_path, workspace_root, host_override, port_override)
    try:
        return forward_once_fn(request, host, port, timeout_sec)
    except (DaemonForwardError, OSError, TimeoutError) as first_exc:
        if not auto_start_on_failure:
            raise
        if not is_retryable_error(first_exc):
            raise
        started = start_daemon_fn(db_path, workspace_root)
        if not started:
            raise
        last_exc: BaseException | None = None
        for _ in range(retry_max):
            host_retry, port_retry = resolve_target_fn(db_path, workspace_root, host_override, port_override)
            try:
                return forward_once_fn(request, host_retry, port_retry, timeout_sec)
            except (DaemonForwardError, OSError, TimeoutError) as retry_exc:
                last_exc = retry_exc
                if not is_retryable_error(retry_exc):
                    raise
                time.sleep(retry_wait_sec)
        if last_exc is not None:
            raise last_exc
        raise


def resolve_post_start_retry_policy(request: dict[str, object]) -> tuple[int, float]:
    """요청 메서드에 따라 post-start 재시도 정책을 반환한다."""
    method_raw = request.get("method")
    method = str(method_raw).strip().lower() if isinstance(method_raw, str) else ""
    if method == "initialize":
        return INITIALIZE_POST_START_RETRY_MAX, INITIALIZE_POST_START_RETRY_WAIT_SEC
    return POST_START_RETRY_MAX, POST_START_RETRY_WAIT_SEC


def is_retryable_error(exc: BaseException) -> bool:
    """재시도 가능한 예외 타입인지 판정한다."""
    return isinstance(exc, (OSError, TimeoutError, DaemonForwardError))


def build_forward_error_message(exc: BaseException) -> str:
    """forward 오류를 명시적 코드 접두사로 표준화한다."""
    raw_message = str(exc)
    if raw_message.startswith("ERR_"):
        return raw_message
    if isinstance(exc, TimeoutError):
        detail_code = "ERR_DAEMON_TIMEOUT"
    elif isinstance(exc, OSError):
        detail_code = "ERR_DAEMON_UNAVAILABLE"
    elif isinstance(exc, DaemonForwardError):
        detail_code = "ERR_DAEMON_PROTOCOL"
    else:
        detail_code = "ERR_DAEMON_FORWARD_UNKNOWN"
    return f"ERR_DAEMON_FORWARD_FAILED: {detail_code}: {exc}"


def default_start_daemon(db_path: Path, workspace_root: str | None) -> bool:
    """기본 daemon 자동 기동/attach 전략을 수행한다."""
    _ = workspace_root
    config = replace(AppConfig.default(), db_path=db_path)
    runtime_repo = RuntimeRepository(db_path)
    workspace_repo = WorkspaceRepository(db_path)
    registry_repo = DaemonRegistryRepository(db_path)
    service = DaemonService(
        config=config,
        runtime_repo=runtime_repo,
        workspace_repo=workspace_repo,
        registry_repo=registry_repo,
    )
    try:
        runtime = service.ensure_running(run_mode=config.run_mode)
        return wait_for_daemon_ready(
            host=runtime.host,
            port=runtime.port,
            timeout_sec=DAEMON_START_READY_TIMEOUT_SEC,
            poll_sec=DAEMON_START_READY_POLL_SEC,
        )
    except DaemonError:
        return False


def wait_for_daemon_ready(host: str, port: int, timeout_sec: float, poll_sec: float) -> bool:
    """daemon TCP endpoint가 연결 가능한 상태가 될 때까지 대기한다."""
    deadline = time.monotonic() + timeout_sec
    while time.monotonic() < deadline:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(poll_sec)
            if sock.connect_ex((host, port)) == 0:
                return True
        time.sleep(poll_sec)
    return False
